# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"
__all__ = ["LocalApp"]

import abc
import copy
from os import chdir, getcwd, remove
from .application import Application, AppState, AppStateError, requires_state
from subprocess import Popen, PIPE, SubprocessError, TimeoutExpired

class LocalApp(Application, metaclass=abc.ABCMeta):
    """
    The base class for all locally installed applications, that are used
    via the command line.
    
    Internally this creates a :class:`Popen` instance, which handles
    the execution.
    
    Parameters
    ----------
    bin_path : str
        Path of the application represented by this class.
    """
    
    def __init__(self, bin_path):
        super().__init__()
        self._bin_path = bin_path
        self._arguments = []
        self._options = []
        self._exec_dir = getcwd()
        self._process = None
        self._command = None
    
    @requires_state(AppState.CREATED)
    def set_arguments(self, arguments):
        """
        Set command line arguments for the application run.
        
        PROTECTED: Do not call from outside.
        
        Parameters
        ----------
        arguments : list of str
            A list of strings representing the command line options.
        """
        self._arguments = copy.copy(arguments)
    
    @requires_state(AppState.CREATED)
    def add_additional_options(self, options):
        """
        Add additional options for the command line program.
        These options are put before the arguments automatically
        determined by the respective :class:`LocalApp` subclass.

        This method is focused on advanced users, who have knowledge on
        the available options of the command line program and the
        options already used by the :class:`LocalApp` subclasses.
        Ignoring the already used options may result in conflicting
        CLI arguments and potential unexpected results.
        It is recommended to use this method only, when the respective
        :class:`LocalApp` subclass does not provide a method to set the
        desired option.
        
        Parameters
        ----------
        options : list of str
            A list of strings representing the command line options.
        
        Notes
        -----
        In order to see which options the command line execution used,
        try the :meth:`get_command()` method.

        Examples
        --------

        >>> seq1 = ProteinSequence("BIQTITE")
        >>> seq2 = ProteinSequence("TITANITE")
        >>> seq3 = ProteinSequence("BISMITE")
        >>> seq4 = ProteinSequence("IQLITE")
        >>> # Run application without additional arguments
        >>> app = ClustalOmegaApp([seq1, seq2, seq3, seq4])
        >>> app.start()
        >>> app.join()
        >>> print(app.get_command())
        clustalo --in ...fa --out ...fa --force --output-order=tree-order --seqtype Protein --guidetree-out ...tree
        >>> # Run application with additional argument
        >>> app = ClustalOmegaApp([seq1, seq2, seq3, seq4])
        >>> app.add_additional_options(["--full"])
        >>> app.start()
        >>> app.join()
        >>> print(app.get_command())
        clustalo --full --in ...fa --out ...fa --force --output-order=tree-order --seqtype Protein --guidetree-out ...tree
        """
        self._options += options
    
    @requires_state(
        AppState.RUNNING | \
        AppState.CANCELLED | \
        AppState.FINISHED | \
        AppState.JOINED
    )
    def get_command(self):
        """
        Get the executed command.

        Cannot be called until the application has been started.
        
        Returns
        -------
        command : str
            The executed command.
        
        Examples
        --------
        
        >>> seq1 = ProteinSequence("BIQTITE")
        >>> seq2 = ProteinSequence("TITANITE")
        >>> seq3 = ProteinSequence("BISMITE")
        >>> seq4 = ProteinSequence("IQLITE")
        >>> app = ClustalOmegaApp([seq1, seq2, seq3, seq4])
        >>> app.start()
        >>> print(app.get_command())
        clustalo --in ...fa --out ...fa --force --output-order=tree-order --seqtype Protein --guidetree-out ...tree
        """
        return " ".join(self._command)

    
    @requires_state(AppState.CREATED)
    def set_exec_dir(self, exec_dir):
        """
        Set the directory where the application should be executed.
        If not set, it will be executed in the working directory at the
        time the application was created. 
        
        PROTECTED: Do not call from outside.
        
        Parameters
        ----------
        exec_dir : str
            The execution directory.
        """
        self._exec_dir = exec_dir
    
    @requires_state(AppState.RUNNING | AppState.FINISHED)
    def get_process(self):
        """
        Get the `Popen` instance.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        process : Popen
            The `Popen` instance
        """
        return self._process
    
    @requires_state(AppState.FINISHED | AppState.JOINED)
    def get_exit_code(self):
        """
        Get the exit code of the process.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        code : int
            The exit code.
        """
        return self._process.returncode
    
    @requires_state(AppState.FINISHED | AppState.JOINED)
    def get_stdout(self):
        """
        Get the STDOUT pipe content of the process.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        stdout : str
            The standard outpout.
        """
        return self._stdout
    
    @requires_state(AppState.FINISHED | AppState.JOINED)
    def get_stderr(self):
        """
        Get the STDERR pipe content of the process.
        
        PROTECTED: Do not call from outside.
        
        Returns
        -------
        stdout : str
            The standard error.
        """
        return self._stderr

    def run(self):
        cwd = getcwd()
        chdir(self._exec_dir) 
        self._command = [self._bin_path] + self._options + self._arguments
        self._process = Popen(
            self._command, stdout=PIPE, stderr=PIPE, encoding="UTF-8"
        )
        chdir(cwd)
    
    def is_finished(self):
        code = self._process.poll()
        if code == None:
            return False
        else:
            self._stdout, self._stderr = self._process.communicate()
            return True
    
    @requires_state(AppState.RUNNING | AppState.FINISHED)
    def join(self, timeout=None):
        # Override method as repetitive calls of 'is_finished()'
        # are not necessary as 'communicate()' already waits for the
        # finished application
        try:
            self._stdout, self._stderr = self._process.communicate(
                timeout=timeout
            )
        except TimeoutExpired:
            self.cancel()
            raise TimeoutError(
                f"The application expired its timeout ({timeout:.1f} s)"
            )
        self._state = AppState.FINISHED
        
        try:
            self.evaluate()
        except AppStateError:
            raise
        except:
            self._state = AppState.CANCELLED
            raise
        else:
            self._state = AppState.JOINED
        self.clean_up()
    
    
    def wait_interval(self):
        # Not used in this implementation of 'join()'
        raise NotImplementedError()
    
    def evaluate(self):
        super().evaluate()
        # Check if applicaion terminated correctly
        exit_code = self.get_exit_code()
        if exit_code != 0:
            err_msg = self.get_stderr().replace("\n", " ")
            raise SubprocessError(
                f"'{self._bin_path}' returned with exit code {exit_code}: "
                f"{err_msg}"
            )
    
    def clean_up(self):
        if self.get_app_state() == AppState.CANCELLED:
            self._process.kill()


def cleanup_tempfile(temp_file):
    """
    Close a :class:`NamedTemporaryFile` and delete it manually,
    if `delete` is set to ``False``.
    This function is a small helper function intended for usage in
    `LocalApp` subclasses. 

    The manual deletion is necessary, as Windows does not allow to open
    a :class:`NamedTemporaryFile` as second time
    (e.g. by the file name), if `delete` is set to ``True``.

    Parameters
    ----------
    temp_file : NamedTemporaryFile
        The temporary file to be closed and deleted.
    """
    temp_file.close()
    if not temp_file.delete:
        remove(temp_file.name)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing trajectories in the
*AMBER* NetCDF format.
"""

__name__ = "biotite.structure.io.netcdf"
__author__ = "Patrick Kunzmann"

from .file import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Functions for converting an annotation from/to a GenBank file.
"""

__name__ = "biotite.sequence.io.genbank"
__author__ = "Patrick Kunzmann"
__all__ = ["get_annotation", "set_annotation"]

import re
import warnings
from ....file import InvalidFileError
from ...annotation import Annotation, Feature, Location
from .file import GenBankFile


_KEY_START = 5
_QUAL_START = 21


def get_annotation(gb_file, include_only=None):
    """
    Get the sequence annotation from the *FEATURES* field of a
    GenBank file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *FEATURES* field from.
    include_only : iterable object of str, optional
        List of names of feature keys, which should included
        in the annotation. By default all features are included.
    
    Returns
    ----------
    annotation : Annotation
        Sequence annotation from the file.
    """
    fields = gb_file.get_fields("FEATURES")
    if len(fields) == 0:
        raise InvalidFileError("File has no 'FEATURES' field")
    if len(fields) > 1:
        raise InvalidFileError("File has multiple 'FEATURES' fields")
    lines, _ = fields[0]
    

    ### Parse all lines to create an index of features,
    # i.e. pairs of the feature key
    # and the text belonging to the respective feature 
    feature_list = []
    feature_key = None
    feature_value = ""
    for line in lines:
        # Check if line contains feature key
        if line[_KEY_START] != " ":
            if feature_key is not None:
                # Store old feature key and value
                feature_list.append((feature_key, feature_value))
            # Track new key
            feature_key = line[_KEY_START : _QUAL_START-1].strip()
            feature_value = ""
        feature_value += line[_QUAL_START:] + " "
    # Store last feature key and value (loop already exited)
    feature_list.append((feature_key, feature_value))
    

    ### Process only relevant features and put them into an Annotation
    annotation = Annotation()
    # Regex to separate qualifiers from each other
    regex = re.compile(r"""(".*?"|/.*?=)""")
    for key, val in feature_list:
        if include_only is None or key in include_only:
            qual_dict = {}
            
            # Split feature definition into parts
            # e.g.
            #
            # 1..12
            # /gene="abcA"
            # /product="AbcA"
            #
            # becomes
            #
            # ['1..12', '/gene=', '"abcA"', '/product=', '"AbcA"']
            qualifier_parts = [s.strip() for s in regex.split(val)]
            # Remove empty qualifier parts
            qualifier_parts = [s for s in qualifier_parts if s]
            # First part is location identifier
            loc_string = qualifier_parts.pop(0).strip()
            try:
                locs = _parse_locs(loc_string)
            except:
                warnings.warn(
                    f"'{loc_string}' is an unsupported location identifier, "
                    f"skipping feature"
                )
                continue

            # The other parts are pairwise qualifier keys and values
            qual_key = None
            qual_val = None
            for part in qualifier_parts:
                if qual_key is None:
                    # This is a qualifier key
                    # When the feature contains qualifiers without
                    # value, e.g. '/pseudo'
                    # The part may contain multiple keys, e.g.
                    #
                    # '/pseudo /gene='
                    #
                    # -> split at whitespaces,
                    # as keys do not contain whitespaces
                    for subpart in part.split():
                        if not "=" in subpart:
                            # Qualifier without value, e.g. '/pseudo'
                            # -> store immediately
                            # Remove "/" -> subpart[1:]
                            qual_key = subpart[1:]
                            _set_qual(qual_dict, qual_key, None)
                            qual_key = None
                        else:
                            # Regular qualifier
                            # -> store key in variable and wait for
                            # next qualifier part to set the value
                            # Remove '/' and '=' -> subpart[1:-1]
                            qual_key = subpart[1:-1]
                else:
                    # This is a qualifier value
                    # -> remove potential quotes
                    if part[0] == '"':
                        qual_val = part[1:-1]
                    else:
                        qual_val = part
                    # Store qualifier pair
                    _set_qual(qual_dict, qual_key, qual_val)
                    qual_key = None
                    qual_val = None
            
            annotation.add_feature(Feature(key, locs, qual_dict))
    
    return annotation


def _parse_locs(loc_str):
    locs = []
    if loc_str.startswith(("join", "order")):
        str_list = loc_str[loc_str.index("(")+1:loc_str.rindex(")")].split(",")
        for s in str_list:
            locs.extend(_parse_locs(s))
    elif loc_str.startswith("complement"):
        compl_str = loc_str[loc_str.index("(")+1:loc_str.rindex(")")]
        compl_locs = [
            Location(loc.first, loc.last, Location.Strand.REVERSE, loc.defect) 
            for loc in _parse_locs(compl_str)
        ]
        locs.extend(compl_locs)
    else:
        locs = [_parse_single_loc(loc_str)]
    return locs


def _parse_single_loc(loc_str):
    if ".." in loc_str:
        split_char = ".."
        defect = Location.Defect.NONE
    elif "." in loc_str:
        split_char = "."
        defect = Location.Defect.UNK_LOC
    elif "^" in loc_str:
        split_char = "^"
        loc_str_split = loc_str.split("..")
        defect = Location.Defect.BETWEEN
    else:
        # Parse single location
        defect = Location.Defect.NONE
        if loc_str[0] == "<":
            loc_str = loc_str[1:]
            defect |= Location.Defect.BEYOND_LEFT
        elif loc_str[0] == ">":
            loc_str = loc_str[1:]
            defect |= Location.Defect.BEYOND_RIGHT
        first_and_last = int(loc_str)
        return Location(first_and_last, first_and_last, defect=defect)
    # Parse location range
    loc_str_split = loc_str.split(split_char)
    first_str = loc_str_split[0]
    last_str = loc_str_split[1]
    # Parse Defects
    if first_str[0] == "<":
        first = int(first_str[1:])
        defect |= Location.Defect.BEYOND_LEFT
    else:
        first = int(first_str)
    if last_str[0] == ">":
        last = int(last_str[1:])
        defect |= Location.Defect.BEYOND_RIGHT
    else:
        last = int(last_str)
    return Location(first, last, defect=defect)


def _set_qual(qual_dict, key, val):
    """
    Set a mapping key to val in the dictionary.
    If the key already exists in the dictionary, append the value (str)
    to the existing value, separated by a line break.
    """
    if key in qual_dict:
        qual_dict[key] += "\n" + val
    else:
        qual_dict[key] = val




def set_annotation(gb_file, annotation):
    """
    Set the *FEATURES* field of a GenBank file with an annotation.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to be edited.
    annotation : Annotation
        The annotation that is put into the GenBank file.
    """
    lines = []
    for feature in sorted(annotation):
        line = " " * _KEY_START
        line += feature.key.ljust(_QUAL_START - _KEY_START)
        line += _convert_to_loc_string(feature.locs)
        lines.append(line)
        for key, values in feature.qual.items():
            if values is None:
                line = " " * _QUAL_START
                line +=  f'/{key}'
                lines.append(line)
            else:
                for val in values.split("\n"):
                    line = " " * _QUAL_START
                    line +=  f'/{key}="{val}"'
                    lines.append(line)
    gb_file.set_field("FEATURES", lines)


def _convert_to_loc_string(locs):
    """
    Create GenBank comptabile location strings from a list of :class:`Location`
    objects.
    """
    if len(locs) == 1:
        loc = list(locs)[0]
        loc_first_str = str(loc.first)
        loc_last_str  = str(loc.last)
        if loc.defect & Location.Defect.BEYOND_LEFT:
            loc_first_str = "<" + loc_first_str
        if loc.defect & Location.Defect.BEYOND_RIGHT:
            loc_last_str  = ">" + loc_last_str
        if loc.first == loc.last:
            loc_string = loc_first_str
        elif loc.defect & Location.Defect.UNK_LOC:
            loc_string = loc_first_str + "." + loc_last_str
        elif loc.defect & Location.Defect.BETWEEN:
            loc_string = loc_first_str + "^" + loc_last_str
        else:
            loc_string = loc_first_str + ".." + loc_last_str
        if loc.strand == Location.Strand.REVERSE:
            loc_string = f"complement({loc_string})"
    else:
        loc_string = ",".join(
            [_convert_to_loc_string([loc]) for loc in locs]
        )
        loc_string = f"join({loc_string})"
    return loc_string



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for obtaining all kinds of chemical information about atoms
and residues, including masses, radii, bonds, etc.

Most information is extracted from the chemical compound dictionary
of the
`wwPDB <ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif>`_
via tools from the
`biotite-util <https://github.com/biotite-dev/biotite-util>`_
repository.
"""

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"

from .atoms import *
from .bonds import *
from .masses import *
from .misc import *
from .radii import *
from .standardize import *
from .nucleotides import *



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.npz"
__author__ = "Patrick Kunzmann"
__all__ = ["NpzFile"]

import numpy as np
from ...atoms import Atom, AtomArray, AtomArrayStack
from ...bonds import BondList
from ....file import File, is_binary


class NpzFile(File):
    r"""
    This class represents a NPZ file, the preferable format for
    Biotite internal structure storage. 
    
    Internally the this class writes/reads all attribute arrays of an
    :class:`AtomArray` or :class:`AtomArrayStack` using the *NumPy*
    :func:`save()`/:func:`load()`
    method. This format offers the fastest I/O operations and completely
    preserves the content all atom annotation arrays.
    
    Examples
    --------
    Load a \\*.npz file, modify the structure and save the new
    structure into a new file:
    
    >>> import os.path
    >>> file = NpzFile.read(os.path.join(path_to_structures, "1l2y.npz"))
    >>> array_stack = file.get_structure()
    >>> array_stack_mod = rotate(array_stack, [1,2,3])
    >>> file = NpzFile()
    >>> file.set_structure(array_stack_mod)
    >>> file.write(os.path.join(path_to_directory, "1l2y_mod.npz"))
    
    """
    
    def __init__(self):
        super().__init__()
        self._data_dict = None
    
    def __copy_fill__(self, clone):
        super().__copy_fill__(clone)
        if self._data_dict is not None:
            for key, value in self._data_dict.items():
                clone._data_dict[key] = np.copy(value)
    
    @classmethod
    def read(cls, file):
        """
        Read a NPZ file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Returns
        -------
        file_object : NPZFile
            The parsed file.
        """
        npz_file = NpzFile()
        # File name
        if isinstance(file, str):
            with open(file, "rb") as f:
                npz_file._data_dict = dict(np.load(f, allow_pickle=False))
        # File object
        else:
            if not is_binary(file):
                raise TypeError("A file opened in 'binary' mode is required")
            npz_file._data_dict = dict(np.load(file, allow_pickle=False))
        return npz_file
                
    def write(self, file):
        """
        Write a NPZ file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively, a file path can be supplied.
        """
        if isinstance(file, str):
            with open(file, "wb") as f:
                np.savez(f, **self._data_dict)
        else:
            if not is_binary(file):
                raise TypeError("A file opened in 'binary' mode is required")
            np.savez(file, **self._data_dict)
    
    def get_structure(self):
        """
        Get an :class:`AtomArray` or :class:`AtomArrayStack` from the
        file.
        
        If this method returns an array or stack depends on which type
        of object was used when the file was written.
        
        Returns
        -------
        array : AtomArray or AtomArrayStack
            The array or stack contained in this file.
        """
        if self._data_dict is None:
            raise ValueError("The structure of this file "
                             "has not been loaded or set yet")
        coord = self._data_dict["coord"]
        # The type of the structure is determined by the dimensionality
        # of the 'coord' field
        if len(coord.shape) == 3:
            array = AtomArrayStack(coord.shape[0], coord.shape[1])
        else:
            array = AtomArray(coord.shape[0])
        
        for key, value in self._data_dict.items():
            if key == "coord":
                array.coord = value
            elif key == "bonds":
                array.bonds = BondList(array.array_length(), value)
            elif key == "box":
                array.box = value
            else:
                array.set_annotation(key, value)
        return array
        
    def set_structure(self, array):
        """
        Set the :class:`AtomArray` or :class:`AtomArrayStack` for the
        file.
        
        Parameters
        ----------
        array : AtomArray or AtomArrayStack
            The array or stack to be saved into this file.
        """
        self._data_dict = {}
        self._data_dict["coord"] = np.copy(array.coord)
        if array.bonds is not None:
            self._data_dict["bonds"] = array.bonds.as_array()
        if array.box is not None:
            self._data_dict["box"] = np.copy(array.box)
        for annot in array.get_annotation_categories():
            self._data_dict[annot] = np.copy(array.get_annotation(annot))


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
from biotite.structure import Atom
import numpy as np
import pytest

@pytest.fixture
def array():
    atom_list = []
    atom_list.append(Atom([0.5, 2.5, 1.0]))
    atom_list.append(Atom([0.5, 2.7, 1.0]))
    atom_list.append(Atom([1.5, 1.5, 1.0]))
    atom_list.append(Atom([2.5, 0.5, 1.0]))
    atom_list.append(Atom([2.5, 0.7, 1.0]))
    atom_list.append(Atom([2.5, 0.5, 1.1]))
    return struc.array(atom_list)

@pytest.fixture
def stack(array):
    return struc.stack([array, array.copy()])



def test_density(array, stack):
    density, (x, y, z) = struc.density(array)
    assert np.array_equal(x, [0.5, 1.5, 2.5])
    assert np.array_equal(y, [0.5, 1.5, 2.5, 3.5])
    assert np.array_equal(z, [1.0, 2.0])
    assert density.sum() == 6
    assert density[0,2] == 2
    assert density[1,0] == 3
    assert density[1,1] == 1

    density, (x, y, z) = struc.density(stack)
    assert np.array_equal(x, [0.5, 1.5, 2.5])
    assert np.array_equal(y, [0.5, 1.5, 2.5, 3.5])
    assert np.array_equal(z, [1.0, 2.0])
    assert density.sum() == 12
    assert density[0,2] == 4
    assert density[1,0] == 6
    assert density[1,1] == 2

def test_density_with_bins(array):
    bins = np.array([[0, 1, 2, 3],[0, 1, 2, 3],[0, 1, 2, 3]])
    density, (x, y, z) = struc.density(array, bins=bins)
    assert np.array_equal(x, [0,1,2,3])
    assert np.array_equal(y, [0,1,2,3])
    assert np.array_equal(z, [0,1,2,3])
    assert density.sum() == 6
    assert density[0,2,1] == 2
    assert density[1,1,1] == 1
    assert density[2,0,1] == 3

def test_density_with_delta(array):
    density, (x, y, z) = struc.density(array, delta=5.0)
    assert density.shape == (1, 1, 1)
    assert density.sum() == 6
    assert density[0,0,0] == 6

def test_density_normalized(array):
    density, (x, y, z) = struc.density(array, density=True)
    assert np.abs(density.sum() - 1.0) < 0.0001
    assert np.abs(density[0,2] - 2.0/6.0) < 0.0001

def test_density_weights(array, stack):
    # assign weights to coordinates
    atomic_weights = np.arange(0.1, 0.7, 0.1)

    # weights should be applied correctly
    density, (x, y, z) = struc.density(array, weights=atomic_weights)
    assert density.sum() == atomic_weights.sum()
    assert density[0, 2] == atomic_weights[0] + atomic_weights[1]
    assert density[1, 0] == atomic_weights[3:].sum()
    assert density[1,1] == atomic_weights[2]

    # weights should be repeated along stack dimensions and lead to the same
    # result independent of shape
    density, (x, y, z) = struc.density(stack, weights=atomic_weights)
    density2, (x, y, z) = struc.density(stack,
                            weights=np.array([atomic_weights, atomic_weights]))
    assert density.sum() == density2.sum()
    assert density[0,2] == density2[0,2]
    assert density[1,0] == density2[1,0]
    assert density[1,1] == density2[1,1]
    



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align


@pytest.fixture
def kmer_alphabet():
    return align.KmerAlphabet(seq.ProteinSequence.alphabet, 3)


np.random.seed(0)
N = 10
@pytest.mark.parametrize("ref_kmer, threshold", zip(
        np.random.randint(10000,  size=N),
        np.random.randint(-5, 15, size=N)
))
def test_score_threshold_rule(kmer_alphabet, ref_kmer, threshold):
    """
    Test if the similar k-mers given by :class:`ScoreThresholdRule`
    are equal to k-mers generated by a brute-force approach.
    """
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    
    ref_kmer_sequence = seq.ProteinSequence()
    ref_kmer_sequence.code = kmer_alphabet.split(ref_kmer)
    
    ref_sim_kmer_set = set()
    # Iterate through all possible k-mers 
    for kmer in range(len(kmer_alphabet)):
        kmer_sequence = seq.ProteinSequence()
        kmer_sequence.code = kmer_alphabet.split(kmer)
        score = align.align_ungapped(
            ref_kmer_sequence, kmer_sequence, matrix, score_only=True
        )
        # Add k-mer to list if the threshold score is reached
        if score >= threshold:
            ref_sim_kmer_set.add(kmer)
    
    test_rule = align.ScoreThresholdRule(matrix, threshold)
    test_sim_kmer_set = set(test_rule.similar_kmers(kmer_alphabet, ref_kmer))

    assert test_sim_kmer_set == ref_sim_kmer_set


def test_invalid_matrix(kmer_alphabet):
    """
    Expect an error when the k-mer alphabet is incompatible with the
    substitution matrix.
    """
    test_rule = align.ScoreThresholdRule(
        align.SubstitutionMatrix.std_nucleotide_matrix(), 0
    )
    with pytest.raises(ValueError):
        test_rule.similar_kmers(kmer_alphabet, 0)


@pytest.mark.parametrize("invalid_kmer", [-1, 1000000])
def test_invalid_kmer(kmer_alphabet, invalid_kmer):
    """
    Expect an error when an invalid k-mer code is given as input.
    """
    test_rule = align.ScoreThresholdRule(
        align.SubstitutionMatrix.std_protein_matrix(), 0
    )
    with pytest.raises(seq.AlphabetError):
        test_rule.similar_kmers(kmer_alphabet, invalid_kmer)


import itertools
from os.path import join
import numpy as np
import pytest
from biotite.structure.io import load_structure
from biotite.structure.rdf import rdf
from biotite.structure.box import vectors_from_unitcell
from ..util import data_dir, cannot_import


TEST_FILE = join(data_dir("structure"), "waterbox.gro")


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
def test_rdf():
    """ General test to reproduce oxygen RDF for a box of water"""
    test_file = TEST_FILE
    stack = load_structure(test_file)

    # calculate oxygen RDF for water
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 10])
    n_bins = 100
    bins, g_r = rdf(oxygen[:, 0].coord, oxygen, interval=interval,
                    bins=n_bins, periodic=False)

    # Compare with MDTraj
    import mdtraj
    traj = mdtraj.load(TEST_FILE)
    ow = [a.index for a in traj.topology.atoms if a.name == 'O']
    pairs = itertools.product([ow[0]], ow)
    mdt_bins, mdt_g_r = mdtraj.compute_rdf(traj, list(pairs),
                                           r_range=interval/10, n_bins=n_bins,
                                           periodic=False)

    assert np.allclose(bins, mdt_bins*10)
    assert np.allclose(g_r, mdt_g_r, rtol=0.0001)


def test_rdf_bins():
    """ Test if RDF produce correct bin ranges """
    stack = load_structure(TEST_FILE)
    center = stack[:, 0]
    num_bins = 44
    bin_range = (0, 11.7)
    bins, g_r = rdf(center, stack, bins=num_bins, interval=bin_range)
    assert(len(bins) == num_bins)
    assert(bins[0] > bin_range[0])
    assert(bins[1] < bin_range[1])


def test_rdf_with_selection():
    """ Test if the selection argument of rdf function works as expected """
    stack = load_structure(TEST_FILE)

    # calculate oxygen RDF for water with and without a selection
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 10])
    n_bins = 100
    sele = (stack.atom_name == 'OW') & (stack.res_id >= 3)
    bins, g_r = rdf(oxygen[:, 0].coord, stack, selection=sele,
                    interval=interval, bins=n_bins, periodic=False)

    nosel_bins, nosel_g_r = rdf(oxygen[:, 0].coord, oxygen[:, 1:],
                                interval=interval, bins=n_bins, periodic=False)

    assert np.allclose(bins, nosel_bins)
    assert np.allclose(g_r, nosel_g_r)


def test_rdf_atom_argument():
    """ Test if the first argument allows to use AtomArrayStack """
    stack = load_structure(TEST_FILE)

    # calculate oxygen RDF for water with and without a selection
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 10])
    n_bins = 100

    bins, g_r = rdf(oxygen[:, 0], stack, interval=interval,
                    bins=n_bins, periodic=False)

    atom_bins, atoms_g_r = rdf(oxygen[:, 0].coord, stack, interval=interval,
                    bins=n_bins, periodic=False)

    assert np.allclose(g_r, atoms_g_r)


def test_rdf_multiple_center():
    """ Test if the first argument allows to use multiple centers"""
    stack = load_structure(TEST_FILE)

    # calculate oxygen RDF for water with and without a selection
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 10])
    n_bins = 100

    # averaging individual calculations
    bins1, g_r1 = rdf(oxygen[:, 1].coord, oxygen[:, 2:], interval=interval,
                    bins=n_bins, periodic=False)
    bins2, g_r2 = rdf(oxygen[:, 0].coord, oxygen[:, 2:], interval=interval,
                    bins=n_bins, periodic=False)
    mean = np.mean([g_r1, g_r2], axis=0)

    # this should give the same result as averaging for oxygen 0 and 1
    bins, g_r = rdf(oxygen[:, 0:2].coord, oxygen[:, 2:], interval=interval,
                    bins=n_bins, periodic=False)

    assert np.allclose(g_r, mean, rtol=0.0001)


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
def test_rdf_periodic():
    """ Test if the periodic argument gives the correct results"""
    test_file = TEST_FILE
    stack = load_structure(test_file)

    # calculate oxygen RDF for water
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 10])
    n_bins = 100
    bins, g_r = rdf(oxygen[:, 0].coord, oxygen[:, 1:], interval=interval,
                    bins=n_bins, periodic=True)

    # Compare with MDTraj
    import mdtraj
    traj = mdtraj.load(TEST_FILE)
    ow = [a.index for a in traj.topology.atoms if a.name == 'O']
    pairs = itertools.product([ow[0]], ow[1:])
    mdt_bins, mdt_g_r = mdtraj.compute_rdf(traj, list(pairs),
                                           r_range=interval/10, n_bins=n_bins,
                                           periodic=True)

    assert np.allclose(bins, mdt_bins*10)
    assert np.allclose(g_r, mdt_g_r, rtol=0.0001)


def test_rdf_box():
    """ Test correct use of simulation boxes """
    stack = load_structure(TEST_FILE)
    box = vectors_from_unitcell(1, 1, 1, 90, 90, 90)
    box = np.repeat(box[np.newaxis, :, :], len(stack), axis=0)

    # Use box attribute of stack
    rdf(stack[:, 0], stack)

    # Use box attribute and dont fail because stack has no box
    stack.box = None
    rdf(stack[:, 0], stack, box=box)

    # Fail if no box is present
    with pytest.raises(ValueError):
        rdf(stack[:, 0], stack)

    # Fail if box is of wrong size
    with pytest.raises(ValueError):
        rdf(stack[:, 0], stack, box=box[0])


def test_rdf_normalized():
    """ Assert that the RDF tail is normalized to 1"""
    test_file = TEST_FILE
    stack = load_structure(test_file)

    # calculate oxygen RDF for water
    oxygen = stack[:, stack.atom_name == 'OW']
    interval = np.array([0, 5])
    n_bins = 100

    bins, g_r = rdf(oxygen.coord, oxygen, interval=interval,
                    bins=n_bins, periodic=True)
    assert np.allclose(g_r[-10:], np.ones(10), atol=0.1)




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for handling sequences.

A :class:`Sequence` can be seen as a succession of symbols.
The set of symbols, that can occur in a sequence, is defined by an
:class:`Alphabet`.
For example, an unambiguous DNA sequence has an :class:`Alphabet`, that
includes the 4 letters (strings) ``'A'``, ``'C'``, ``'G'`` and ``'T'``.
But furthermore, an :class:`Alphabet` can also contain any immutable and
hashable Python object like :class:`int`, :class:`tuple`, etc.
If a :class:`Sequence` is created with at least a symbol,
that is not in the given :class:`Alphabet`, an :class:`AlphabetError` is
raised.

Internally, a :class:`Sequence` is saved as a *NumPy* :class:`ndarray`
of integer values, where each integer represents a symbol in the
:class:`Alphabet`.
For example, ``'A'``, ``'C'``, ``'G'`` and ``'T'`` would be encoded into
0, 1, 2 and 3, respectively.
These integer values are called *symbol code*, the encoding of an entire
sequence of symbols is called *sequence code*.

The size of the symbol code type in the array is determined by the 
size of the :class:`Alphabet`:
If the :class:`Alphabet` contains 256 symbols or less, one byte is used
per array element, between 257 and 65536 symbols, two bytes are used,
and so on.

This approach has multiple advantages:

    - Wider spectrum of what kind of objects can be represented by
      :class:`Sequence` objects
    - Efficient memory usage and faster calculations due to
      alphabet-tailored *symbol code* type size
    - C-acceleration due to usage of :class:`ndarray` objects
    - Most functions applied on :class:`Sequence` objects are
      indifferent to the actual type of sequence.
    - Symbol codes are directly indices for substitution matrices in
      alignments

The abstract :class:`Sequence` superclass cannot be instantiated
directly, as it does not define an :class:`Alphabet` by itself.
Instead usually the concrete subclasses :class:`NucleotideSequence`
(for DNA and RNA sequences) and :class:`ProteinSequence`
(for amino acid sequences) are used.
These classes have defined alphabets and provide additional sequence
type specific methods.
The class :class:`GeneralSequence` allows the usage of a custom
:class:`Alphabet` without the need to subclass :class:`Sequence`.

Additionally, this subpackage provides support for sequence features,
as used in e.g. GenBank or GFF files.
A :class:`Feature` stores its key name, its qualifiers and locations.
An :class:`Annotation` is a group of multiple :class:`Feataure` objects
and offers convenient location based indexing.
An :class:`AnnotatedSequence` combines an :class:`Annotation` and a
:class:`Sequence`.
"""

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"

from .alphabet import *
from .search import *
from .seqtypes import *
from .sequence import *
from .codon import *
from .annotation import *
from .profile import *



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.pdb"
__author__ = "Patrick Kunzmann, Daniel Bauer, Claude J. Rogers"
__all__ = ["PDBFile"]

import warnings
import numpy as np
from ...atoms import AtomArray, AtomArrayStack
from ...bonds import BondList, connect_via_residue_names
from ...box import vectors_from_unitcell, unitcell_from_vectors
from ....file import TextFile, InvalidFileError
from ..general import _guess_element as guess_element
from ...error import BadStructureError
from ...filter import (
    filter_first_altloc,
    filter_highest_occupancy_altloc,
    filter_solvent,
)
from .hybrid36 import encode_hybrid36, decode_hybrid36, max_hybrid36_number
import copy
from warnings import warn

# slice objects for readability
# ATOM/HETATM
_record = slice(0, 6)
_atom_id = slice(6, 11)
_atom_name = slice(12, 16)
_alt_loc = slice(16, 17)
_res_name = slice(17, 20)
_chain_id = slice(21, 22)
_res_id = slice(22, 26)
_ins_code = slice(26, 27)
_coord_x = slice(30, 38)
_coord_y = slice(38, 46)
_coord_z = slice(46, 54)
_occupancy = slice(54, 60)
_temp_f = slice(60, 66)
_element = slice(76, 78)
_charge = slice(78, 80)
# CRYST1
_a = slice(6, 15)
_b = slice(15, 24)
_c = slice(24, 33)
_alpha = slice(33, 40)
_beta = slice(40, 47)
_gamma = slice(47, 54)


class PDBFile(TextFile):
    r"""
    This class represents a PDB file.
    
    The usage of PDBxFile is encouraged in favor of this class.
    
    This class only provides support for reading/writing the pure atom
    information (*ATOM*, *HETATM*, *MODEL* and *ENDMDL* records). *TER*
    records cannot be written.
    
    See also
    --------
    PDBxFile
    
    Examples
    --------
    Load a `\\*.pdb` file, modify the structure and save the new
    structure into a new file:
    
    >>> import os.path
    >>> file = PDBFile.read(os.path.join(path_to_structures, "1l2y.pdb"))
    >>> array_stack = file.get_structure()
    >>> array_stack_mod = rotate(array_stack, [1,2,3])
    >>> file = PDBFile()
    >>> file.set_structure(array_stack_mod)
    >>> file.write(os.path.join(path_to_directory, "1l2y_mod.pdb"))
    """
    def get_model_count(self):
        """
        Get the number of models contained in the PDB file.

        Returns
        -------
        model_count : int
            The number of models.
        """
        model_count = 0
        for line in self.lines:
            if line.startswith("MODEL"):
                model_count += 1
        
        if model_count == 0:
            # It could be an empty file or a file with a single model,
            # where the 'MODEL' line is missing
            for line in self.lines:
                if line.startswith(("ATOM", "HETATM")):
                    return 1
            return 0
        else:
            return model_count
    
    def get_coord(self, model=None):
        """
        Get only the coordinates of the PDB file.
        
        Parameters
        ----------
        model : int, optional
            If this parameter is given, the function will return a
            2D coordinate array from the atoms corresponding to the
            given model number (starting at 1).
            Negative values are used to index models starting from the
            last model insted of the first model.
            If this parameter is omitted, an 2D coordinate array
            containing all models will be returned, even if
            the structure contains only one model.
        
        Returns
        -------
        coord : ndarray, shape=(m,n,3) or shape=(n,2), dtype=float
            The coordinates read from the ATOM and HETATM records of the
            file.
        
        Notes
        -----
        Note that :func:`get_coord()` may output more coordinates than
        the atom array (stack) from the corresponding
        :func:`get_structure()` call has.
        The reason for this is, that :func:`get_structure()` filters
        *altloc* IDs, while `get_coord()` does not.
        
        Examples
        --------
        Read an :class:`AtomArrayStack` from multiple PDB files, where
        each PDB file contains the same atoms but different positions.
        This is an efficient approach when a trajectory is spread into
        multiple PDB files, as done e.g. by the *Rosetta* modeling
        software. 

        For the purpose of this example, the PDB files are created from
        an existing :class:`AtomArrayStack`.
        
        >>> import os.path
        >>> from tempfile import gettempdir
        >>> file_names = []
        >>> for i in range(atom_array_stack.stack_depth()):
        ...     pdb_file = PDBFile()
        ...     pdb_file.set_structure(atom_array_stack[i])
        ...     file_name = os.path.join(gettempdir(), f"model_{i+1}.pdb")
        ...     pdb_file.write(file_name)
        ...     file_names.append(file_name)
        >>> print(file_names)
        ['...model_1.pdb', '...model_2.pdb', ..., '...model_38.pdb']

        Now the PDB files are used to create an :class:`AtomArrayStack`,
        where each model represents a different model.

        Construct a new :class:`AtomArrayStack` with annotations taken
        from one of the created files used as template and coordinates
        from all of the PDB files.

        >>> template_file = PDBFile.read(file_names[0])
        >>> template = template_file.get_structure()
        >>> coord = []
        >>> for i, file_name in enumerate(file_names):
        ...     pdb_file = PDBFile.read(file_name)
        ...     coord.append(pdb_file.get_coord(model=1))
        >>> new_stack = from_template(template, np.array(coord))

        The newly created :class:`AtomArrayStack` should now be equal to
        the :class:`AtomArrayStack` the PDB files were created from.

        >>> print(np.allclose(new_stack.coord, atom_array_stack.coord))
        True
        """
        # Line indices where a new model starts
        model_start_i = np.array([i for i in range(len(self.lines))
                                  if self.lines[i].startswith("MODEL")],
                                 dtype=int)
        # Line indices with ATOM or HETATM records
        atom_line_i = np.array([i for i in range(len(self.lines)) if
                                self.lines[i].startswith(("ATOM", "HETATM"))],
                                dtype=int)
        # Structures containing only one model may omit MODEL record
        # In these cases model starting index is set to 0
        if len(model_start_i) == 0:
            model_start_i = np.array([0])
        
        if model is None:
            depth = len(model_start_i)
            length = self._get_model_length(model_start_i, atom_line_i)
            coord_i = atom_line_i
        
        else:
            last_model = len(model_start_i)
            if model == 0:
                raise ValueError("The model index must not be 0")
            # Negative models mean index starting from last model
            model = last_model + model + 1 if model < 0 else model

            if model < last_model:
                line_filter = ( ( atom_line_i >= model_start_i[model-1] ) &
                                ( atom_line_i <  model_start_i[model  ] ) )
            elif model == last_model:
                line_filter = (atom_line_i >= model_start_i[model-1])
            else:
                raise ValueError(
                    f"The file has {last_model} models, "
                    f"the given model {model} does not exist"
                )
            coord_i = atom_line_i[line_filter]
            length = len(coord_i)
        
        # Fill in coordinates
        if model is None:
            coord = np.zeros((depth, length, 3), dtype=np.float32)
            m = 0
            i = 0
            for line_i in atom_line_i:
                if m < len(model_start_i)-1 and line_i > model_start_i[m+1]:
                    m += 1
                    i = 0
                line = self.lines[line_i]
                coord[m,i,0] = float(line[30:38])
                coord[m,i,1] = float(line[38:46])
                coord[m,i,2] = float(line[46:54])
                i += 1
            return coord
        
        else:
            coord = np.zeros((length, 3), dtype=np.float32)
            for i, line_i in enumerate(coord_i):
                line = self.lines[line_i]
                coord[i,0] = float(line[30:38])
                coord[i,1] = float(line[38:46])
                coord[i,2] = float(line[46:54])
            return coord

    def get_structure(self, model=None, altloc="first", extra_fields=[],
                      include_bonds=False):
        """
        Get an :class:`AtomArray` or :class:`AtomArrayStack` from the PDB file.
        
        This function parses standard base-10 PDB files as well as
        hybrid-36 PDB.
        
        Parameters
        ----------
        model : int, optional
            If this parameter is given, the function will return an
            :class:`AtomArray` from the atoms corresponding to the given
            model number (starting at 1).
            Negative values are used to index models starting from the
            last model insted of the first model.
            If this parameter is omitted, an :class:`AtomArrayStack`
            containing all models will be returned, even if the
            structure contains only one model.
        altloc : {'first', 'occupancy', 'all'}
            This parameter defines how *altloc* IDs are handled:
                - ``'first'`` - Use atoms that have the first
                  *altloc* ID appearing in a residue.
                - ``'occupancy'`` - Use atoms that have the *altloc* ID
                  with the highest occupancy for a residue.
                - ``'all'`` - Use all atoms.
                  Note that this leads to duplicate atoms.
                  When this option is chosen, the ``altloc_id``
                  annotation array is added to the returned structure.
        extra_fields : list of str, optional
            The strings in the list are optional annotation categories
            that should be stored in the output array or stack.
            These are valid values:
            ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and
            ``'charge'``.
        include_bonds : bool, optional
            If set to true, a :class:`BondList` will be created for the
            resulting :class:`AtomArray` containing the bond information
            from the file.
            All bonds have :attr:`BondType.ANY`, since the PDB format
            does not support bond orders.
        
        Returns
        -------
        array : AtomArray or AtomArrayStack
            The return type depends on the `model` parameter.
        """
        # Line indices where a new model starts
        model_start_i = np.array([i for i in range(len(self.lines))
                                  if self.lines[i].startswith(("MODEL"))],
                                 dtype=int)
        # Line indices with ATOM or HETATM records
        atom_line_i = np.array([i for i in range(len(self.lines)) if
                                self.lines[i].startswith(("ATOM", "HETATM"))],
                               dtype=int)
        # Structures containing only one model may omit MODEL record
        # In these cases model starting index is set to 0
        if len(model_start_i) == 0:
            model_start_i = np.array([0])
        
        if model is None:
            depth = len(model_start_i)
            length = self._get_model_length(model_start_i, atom_line_i)
            array = AtomArrayStack(depth, length)
            # Line indices for annotation determination
            # Annotation is determined from model 1,
            # therefore from ATOM records before second MODEL record
            if len(model_start_i) == 1:
                annot_i = atom_line_i
            else:
                annot_i = atom_line_i[atom_line_i < model_start_i[1]]
            # Line indices for coordinate determination
            coord_i = atom_line_i
        
        else:
            last_model = len(model_start_i)
            if model == 0:
                raise ValueError("The model index must not be 0")
            # Negative models mean index starting from last model
            model = last_model + model + 1 if model < 0 else model

            if model < last_model:
                line_filter = ( ( atom_line_i >= model_start_i[model-1] ) &
                                ( atom_line_i <  model_start_i[model  ] ) )
            elif model == last_model:
                line_filter = (atom_line_i >= model_start_i[model-1])
            else:
                raise ValueError(
                    f"The file has {last_model} models, "
                    f"the given model {model} does not exist"
                )
            annot_i = coord_i = atom_line_i[line_filter]
            array = AtomArray(len(coord_i))
        
        # Create mandatory and optional annotation arrays
        chain_id  = np.zeros(array.array_length(), array.chain_id.dtype)
        res_id    = np.zeros(array.array_length(), array.res_id.dtype)
        ins_code  = np.zeros(array.array_length(), array.ins_code.dtype)
        res_name  = np.zeros(array.array_length(), array.res_name.dtype)
        hetero    = np.zeros(array.array_length(), array.hetero.dtype)
        atom_name = np.zeros(array.array_length(), array.atom_name.dtype)
        element   = np.zeros(array.array_length(), array.element.dtype)
        atom_id_raw = np.zeros(array.array_length(), "U5")
        charge_raw  = np.zeros(array.array_length(), "U2")
        occupancy = np.zeros(array.array_length(), float)
        b_factor  = np.zeros(array.array_length(), float)
        altloc_id = np.zeros(array.array_length(), dtype="U1")

        # Fill annotation array
        # i is index in array, line_i is line index
        for i, line_i in enumerate(annot_i):
            line = self.lines[line_i]
            chain_id[i] = line[_chain_id].upper().strip()
            res_id[i] = decode_hybrid36(line[_res_id])
            ins_code[i] = line[_ins_code].strip()
            res_name[i] = line[_res_name].strip()
            hetero[i] = line[_record] == "HETATM"
            atom_name[i] = line[_atom_name].strip()
            element[i] = line[_element].strip()
            altloc_id[i] = line[_alt_loc]
            atom_id_raw[i] = line[_atom_id]
            charge_raw[i] = line[_charge][::-1]  # turn "1-" into "-1"
            occupancy[i] = float(line[_occupancy].strip())
            b_factor[i] = float(line[_temp_f].strip())
        
        if include_bonds or \
            (extra_fields is not None and "atom_id" in extra_fields):
                # The atom IDs are only required in these two cases
                atom_id = np.array(
                    [decode_hybrid36(raw_id.item()) for raw_id in atom_id_raw],
                    dtype=int
                )
        else:
            atom_id = None
        
        # Add annotation arrays to atom array (stack)
        array.chain_id = chain_id
        array.res_id = res_id
        array.ins_code = ins_code
        array.res_name = res_name
        array.hetero = hetero
        array.atom_name = atom_name
        array.element = element

        for field in (extra_fields if extra_fields is not None else []):
            if field == "atom_id":
                # Copy is necessary to avoid double masking in 
                # later altloc ID filtering
                array.set_annotation("atom_id", atom_id.copy())
            elif field == "charge":
                charge = np.array(charge_raw)
                array.set_annotation("charge", np.where(
                    charge == "  ", "0", charge
                ).astype(int))
            elif field == "occupancy":
                array.set_annotation("occupancy", occupancy)
            elif field == "b_factor":
                array.set_annotation("b_factor", b_factor)
            else:
                raise ValueError(f"Unknown extra field: {field}")

        # Replace empty strings for elements with guessed types
        # This is used e.g. for PDB files created by Gromacs
        if "" in array.element:
            rep_num = 0
            for idx in range(len(array.element)):
                if not array.element[idx]:
                    atom_name = array.atom_name[idx]
                    array.element[idx] = guess_element(atom_name)
                    rep_num += 1
            warn("{} elements were guessed from atom_name.".format(rep_num))
        
        # Fill in coordinates
        if isinstance(array, AtomArray):
            for i, line_i in enumerate(coord_i):
                line = self.lines[line_i]
                array.coord[i, 0] = float(line[_coord_x])
                array.coord[i, 1] = float(line[_coord_y])
                array.coord[i, 2] = float(line[_coord_z])
                
        elif isinstance(array, AtomArrayStack):
            m = 0
            i = 0
            for line_i in atom_line_i:
                if m < len(model_start_i)-1 and line_i > model_start_i[m+1]:
                    m += 1
                    i = 0
                line = self.lines[line_i]
                array.coord[m, i, 0] = float(line[_coord_x])
                array.coord[m, i, 1] = float(line[_coord_y])
                array.coord[m, i, 2] = float(line[_coord_z])
                i += 1

        # Fill in box vectors
        # PDB does not support changing box dimensions. CRYST1 is a one-time
        # record so we can extract it directly
        for line in self.lines:
            if line.startswith("CRYST1"):
                try:
                    len_a = float(line[_a])
                    len_b = float(line[_b])
                    len_c = float(line[_c])
                    alpha = np.deg2rad(float(line[_alpha]))
                    beta = np.deg2rad(float(line[_beta]))
                    gamma = np.deg2rad(float(line[_gamma]))
                    box = vectors_from_unitcell(
                        len_a, len_b, len_c, alpha, beta, gamma
                    )
                except ValueError:
                    # File contains invalid 'CRYST1' record
                    warnings.warn(
                        "File contains invalid 'CRYST1' record, box is ignored"
                    )
                    box = None

                if isinstance(array, AtomArray):
                    array.box = box
                else:
                    array.box = np.repeat(
                        box[np.newaxis, ...], array.stack_depth(), axis=0
                    )
                break  

        # Filter altloc IDs
        if altloc == "occupancy":
            filter = filter_highest_occupancy_altloc(
                array, altloc_id, occupancy
            )
            array = array[..., filter]
            atom_id = atom_id[filter] if atom_id is not None else None
        elif altloc == "first":
            filter = filter_first_altloc(array, altloc_id)
            array = array[..., filter]
            atom_id = atom_id[filter] if atom_id is not None else None
        elif altloc == "all":
            array.set_annotation("altloc_id", altloc_id)
        else:
            raise ValueError(f"'{altloc}' is not a valid 'altloc' option")
        
        # Read bonds
        if include_bonds:
            bond_list = self._get_bonds(atom_id)
            bond_list = bond_list.merge(connect_via_residue_names(
                array,
                # The information for non-hetero residues and water
                # are not part of CONECT records
                (~array.hetero) | filter_solvent(array)
            ))
            # Remove bond order from inter residue bonds for consistency
            bond_list.remove_bond_order()
            array.bonds = bond_list  
        
        return array

    def set_structure(self, array, hybrid36=False):
        """
        Set the :class:`AtomArray` or :class:`AtomArrayStack` for the
        file.
        
        This makes also use of the optional annotation arrays
        ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and ``'charge'``.
        If the atom array (stack) contains the annotation ``'atom_id'``,
        these values will be used for atom numbering instead of
        continuous numbering.
        
        Parameters
        ----------
        array : AtomArray or AtomArrayStack
            The array or stack to be saved into this file. If a stack
            is given, each array in the stack is saved as separate
            model.
        hybrid36: bool, optional
            Defines wether the file should be written in hybrid-36
            format.
        
        Notes
        -----
        If `array` has an associated :class:`BondList`, ``CONECT``
        records are also written for all non-water hetero residues
        and all inter-residue connections.
        """
        natoms = array.array_length()
        annot_categories = array.get_annotation_categories()
        record = np.char.array(np.where(array.hetero, "HETATM", "ATOM"))
        # Check for optional annotation categories
        if "atom_id" in annot_categories:
            atom_id = array.atom_id
        else:
            atom_id = np.arange(1, natoms + 1)
        if "b_factor" in annot_categories:
            b_factor = np.char.array([f"{b:>6.2f}" for b in array.b_factor])
        else:
            b_factor = np.char.array(np.full(natoms, "  0.00", dtype="U6"))
        if "occupancy" in annot_categories:
            occupancy = np.char.array([f"{o:>6.2f}" for o in array.occupancy])
        else:
            occupancy = np.char.array(np.full(natoms, "  1.00", dtype="U6"))
        if "charge" in annot_categories:
            charge = np.char.array(
                [str(np.abs(charge)) + "+" if charge > 0 else
                 (str(np.abs(charge)) + "-" if charge < 0 else "")
                 for charge in array.get_annotation("charge")]
            )
        else:
            charge = np.char.array(np.full(natoms, "  ", dtype="U2"))

        # Do checks on atom array (stack)
        if hybrid36:
            max_atoms = max_hybrid36_number(5)
            max_residues = max_hybrid36_number(4)
        else:
            max_atoms, max_residues = 99999, 9999
        if array.array_length() > max_atoms:
            warn(f"More then {max_atoms:,} atoms per model")
        if (array.res_id > max_residues).any():
            warn(f"Residue IDs exceed {max_residues:,}")
        if np.isnan(array.coord).any():
            raise ValueError("Coordinates contain 'NaN' values")
        if any([len(name) > 1 for name in array.chain_id]):
            raise ValueError("Some chain IDs exceed 1 character")
        if any([len(name) > 3 for name in array.res_name]):
            raise ValueError("Some residue names exceed 3 characters")
        if any([len(name) > 4 for name in array.atom_name]):
            raise ValueError("Some atom names exceed 4 characters")

        if hybrid36:
            pdb_atom_id = np.char.array(
                [encode_hybrid36(i, 5) for i in atom_id]
            )
            pdb_res_id = np.char.array(
                [encode_hybrid36(i, 4) for i in array.res_id]
            )
        else:
            # Atom IDs are supported up to 99999,
            # but negative IDs are also possible
            pdb_atom_id = np.char.array(np.where(
                atom_id > 0,
                ((atom_id - 1) % 99999) + 1,
                atom_id
            ).astype(str))
            # Residue IDs are supported up to 9999,
            # but negative IDs are also possible
            pdb_res_id = np.char.array(np.where(
                array.res_id > 0,
                ((array.res_id - 1) % 9999) + 1,
                array.res_id
            ).astype(str))
        
        names = np.char.array(
            [f" {atm}" if len(elem) == 1 and len(atm) < 4 else atm
             for atm, elem in zip(array.atom_name, array.element)]
        )
        res_names = np.char.array(array.res_name)
        chain_ids = np.char.array(array.chain_id)
        ins_codes = np.char.array(array.ins_code)
        spaces = np.char.array(np.full(natoms, " ", dtype="U1"))
        elements = np.char.array(array.element)

        first_half = (
            record.ljust(6) +
            pdb_atom_id.rjust(5) +
            spaces +
            names.ljust(4) +
            spaces + res_names.rjust(3) + spaces + chain_ids +
            pdb_res_id.rjust(4) + ins_codes.rjust(1)
        )

        second_half = (
            occupancy + b_factor + 10 * spaces +
            elements.rjust(2) + charge.rjust(2)
        )

        coords = array.coord
        if coords.ndim == 2:
            coords = coords[np.newaxis, ...]
        
        self.lines = []
        # Prepend a single CRYST1 record if we have box information
        if array.box is not None:
            box = array.box
            if len(box.shape) == 3:
                box = box[0]
            a, b, c, alpha, beta, gamma = unitcell_from_vectors(box)
            self.lines.append(
                f"CRYST1{a:>9.3f}{b:>9.3f}{c:>9.3f}"
                f"{np.rad2deg(alpha):>7.2f}{np.rad2deg(beta):>7.2f}"
                f"{np.rad2deg(gamma):>7.2f} P 1           1"
            )
        is_stack = coords.shape[0] > 1
        for model_num, coord_i in enumerate(coords, start=1):
            # for an ArrayStack, this is run once
            # only add model lines if is_stack
            if is_stack:
                self.lines.append(f"MODEL     {model_num:4}")
            # Bundle non-coordinate data to simplify iteration
            self.lines.extend(
                [f"{start:27}   {x:>8.3f}{y:>8.3f}{z:>8.3f}{end:26}"
                 for start, (x, y, z), end in
                 zip(first_half, coord_i, second_half)]
            )
            if is_stack:
                self.lines.append("ENDMDL")
        
        # Add CONECT records if bonds are present
        if array.bonds is not None:
            # Only non-water hetero records and connections between
            # residues are added to the records
            hetero_indices = np.where(array.hetero & ~filter_solvent(array))[0]
            bond_array = array.bonds.as_array()
            bond_array = bond_array[
                np.isin(bond_array[:,0], hetero_indices) |
                np.isin(bond_array[:,1], hetero_indices) |
                (array.res_id  [bond_array[:,0]] != array.res_id  [bond_array[:,1]]) |
                (array.chain_id[bond_array[:,0]] != array.chain_id[bond_array[:,1]])
            ]
            self._set_bonds(
                BondList(array.array_length(), bond_array), atom_id
            )

    def _get_model_length(self, model_start_i, atom_line_i):
        """
        Determine length of models and check that all models
        have equal length.
        """
        n_models = len(model_start_i)
        length = None
        for model_i in range(len(model_start_i)):
            model_start = model_start_i[model_i]
            model_stop = model_start_i[model_i+1] if model_i+1 < n_models \
                            else len(self.lines)
            model_length = np.count_nonzero(
                (atom_line_i >= model_start) & (atom_line_i < model_stop)
            )
            if length is None:
                length = model_length
            if model_length != length:
                raise InvalidFileError(
                    f"Model {model_i+1} has {model_length} atoms, "
                    f"but model 1 has {length} atoms, must be equal"
                )
        return length

    def _get_bonds(self, atom_ids):
        conect_lines = [line for line in self.lines
                        if line.startswith("CONECT")]
        
        # Mapping from atom ids to indices in an AtomArray
        atom_id_to_index = np.zeros(atom_ids[-1]+1, dtype=int)
        for i, id in enumerate(atom_ids):
           atom_id_to_index[id] = i

        bonds = []
        for line in conect_lines:
            center_id = atom_id_to_index[int(line[6 : 11])]
            for i in range(11, 31, 5):
                id_string = line[i : i+5]
                try:
                    id = atom_id_to_index[int(id_string)]
                except ValueError:
                    # String is empty -> no further IDs
                    break
                bonds.append((center_id, id))
        
        # The length of the 'atom_ids' array
        # is equal to the length of the AtomArray
        return BondList(len(atom_ids), np.array(bonds, dtype=np.uint32))

    def _set_bonds(self, bond_list, atom_ids):
        # Bond type is unused since PDB does not support bond orders
        bonds, _ = bond_list.get_all_bonds()

        for center_i, bonded_indices in enumerate(bonds):
            n_added = 0
            for bonded_i in bonded_indices:
                if bonded_i == -1:
                    # Reached padding values
                    break
                if n_added == 0:
                    # Add new record
                    line = f"CONECT{atom_ids[center_i]:>5d}"
                line += f"{atom_ids[bonded_i]:>5d}"
                n_added += 1
                if n_added == 4:
                    # Only a maximum of 4 bond partners can be put
                    # into a single line
                    # If there are more, use an extra record
                    n_added = 0
                    self.lines.append(line)
            if n_added > 0:
                self.lines.append(line)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.muscle"
__author__ = "Patrick Kunzmann"
__all__ = ["Muscle5App"]

import numbers
import warnings
from tempfile import NamedTemporaryFile
from ..localapp import cleanup_tempfile
from ..msaapp import MSAApp
from ..application import AppState, VersionError, requires_state
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.align.matrix import SubstitutionMatrix
from ...sequence.align.alignment import Alignment
from ...sequence.phylo.tree import Tree
from .app3 import get_version


class Muscle5App(MSAApp):
    """
    Perform a multiple sequence alignment using MUSCLE version 5.
    
    Parameters
    ----------
    sequences : list of Sequence
        The sequences to be aligned.
    bin_path : str, optional
        Path of the MUSCLE binary.
    
    See also
    ---------
    MuscleApp

    Notes
    -----
    Alignment ensemble generation is not supported, yet.
    
    Examples
    --------

    >>> seq1 = ProteinSequence("BIQTITE")
    >>> seq2 = ProteinSequence("TITANITE")
    >>> seq3 = ProteinSequence("BISMITE")
    >>> seq4 = ProteinSequence("IQLITE")
    >>> app = Muscle5App([seq1, seq2, seq3, seq4])
    >>> app.start()
    >>> app.join()
    >>> alignment = app.get_alignment()
    >>> print(alignment)
    BI-QTITE
    TITANITE
    BI-SMITE
    -I-QLITE
    """
    
    def __init__(self, sequences, bin_path="muscle"):
        major_version = get_version(bin_path)[0]
        if major_version < 5:
            raise VersionError(
                f"At least Muscle 5 is required, got version {major_version}"
            )
        
        super().__init__(sequences, bin_path)
        self._mode = "align"
        self._consiters = None
        self._refineiters = None
        self._n_threads = None

    @requires_state(AppState.CREATED)
    def set_iterations(self, consistency=None, refinement=None):
        """
        Set the number of iterations for the alignment algorithm.

        Parameters
        ----------
        consistency : int, optional
            The number of consistency iterations.
        refinement : int, optional
            The number of refinement iterations.
        """
        if consistency is not None:
            self._consiters = consistency
        if refinement is not None:
            self._refineiters = refinement
    
    @requires_state(AppState.CREATED)
    def set_thread_number(self, number):
        """
        Set the number of threads for the alignment run.

        Parameters
        ----------
        number : int, optional
            The number of threads.
        """
        self._n_threads = number

    @requires_state(AppState.CREATED)
    def use_super5(self):
        """
        Use the *Super5* algorithm for the alignment run.
        """
        self._mode = "super5"

    def run(self):
        args = [
            f"-{self._mode}",
            self.get_input_file_path(),
            "-output", self.get_output_file_path(),
        ]
        if self.get_seqtype() == "protein":
            args += ["-amino"]
        else:
            args += ["-nt"]
        if self._n_threads is not None:
             args += ["-threads", str(self._n_threads)]
        if self._consiters is not None:
             args += ["-consiters", str(self._consiters)]
        if self._refineiters is not None:
             args += ["-refineiters", str(self._refineiters)]
        self.set_arguments(args)
        super().run()
    
    def clean_up(self):
        super().clean_up()
    
    @staticmethod
    def supports_nucleotide():
        return True
    
    @staticmethod
    def supports_protein():
        return True
    
    @staticmethod
    def supports_custom_nucleotide_matrix():
        return False
    
    @staticmethod
    def supports_custom_protein_matrix():
        return False
    
    @classmethod
    def align(cls, sequences, bin_path=None):
        """
        Perform a multiple sequence alignment.
        
        This is a convenience function, that wraps the :class:`Muscle5App`
        execution.
        
        Parameters
        ----------
        sequences : iterable object of Sequence
            The sequences to be aligned
        bin_path : str, optional
            Path of the MSA software binary. By default, the default path
            will be used.
        
        Returns
        -------
        alignment : Alignment
            The global multiple sequence alignment.
        """
        app = cls(sequences, bin_path)
        app.start()
        app.join()
        return app.get_alignment()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.pdbx"
__author__ = "Fabrice Allain, Patrick Kunzmann"
__all__ = [
    "get_sequence",
    "get_model_count",
    "get_structure",
    "set_structure",
    "list_assemblies",
    "get_assembly",
]

import itertools
import warnings
from collections import OrderedDict
import numpy as np
from ....file import InvalidFileError
from ....sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...atoms import AtomArray, AtomArrayStack, repeat
from ...box import unitcell_from_vectors, vectors_from_unitcell
from ...filter import filter_first_altloc, filter_highest_occupancy_altloc
from ...util import matrix_rotate

_proteinseq_type_list = ["polypeptide(D)", "polypeptide(L)"]
_nucleotideseq_type_list = [
    "polydeoxyribonucleotide",
    "polyribonucleotide",
    "polydeoxyribonucleotide/polyribonucleotide hybrid",
]
_other_type_list = [
    "cyclic-pseudo-peptide",
    "other",
    "peptide nucleic acid",
    "polysaccharide(D)",
    "polysaccharide(L)",
]


def get_sequence(pdbx_file, data_block=None):
    """
    Get the protein and nucleotide sequences from the
    ``entity_poly.pdbx_seq_one_letter_code_can`` entry.

    Supported polymer types (``_entity_poly.type``) are:
    ``'polypeptide(D)'``, ``'polypeptide(L)'``,
    ``'polydeoxyribonucleotide'``, ``'polyribonucleotide'`` and
    ``'polydeoxyribonucleotide/polyribonucleotide hybrid'``.
    Uracil is converted to Thymine.

    Parameters
    ----------
    pdbx_file : PDBxFile
        The file object.
    data_block : string, optional
        The name of the data block. Default is the first
        (and most times only) data block of the file.

    Returns
    -------
    sequences : list of Sequence
        The protein and nucleotide sequences for each entity
        (equivalent to chains in most cases).
    """
    poly_dict = pdbx_file.get_category("entity_poly", data_block)
    seq_string = poly_dict["pdbx_seq_one_letter_code_can"]
    seq_type = poly_dict["type"]
    sequences = []
    if isinstance(seq_string, np.ndarray):
        for string, stype in zip(seq_string, seq_type):
            sequence = _convert_string_to_sequence(string, stype)
            if sequence is not None:
                sequences.append(sequence)
    else:
        sequences.append(_convert_string_to_sequence(seq_string, seq_type))
    return sequences


def get_model_count(file, data_block=None):
    """
    Get the number of models contained in a :class:`PDBxFile`.

    Parameters
    ----------
    file : PDBxFile
        The file object.
    data_block : str, optional
        The name of the data block. Default is the first
        (and most times only) data block of the file.

    Returns
    -------
    model_count : int
        The number of models.
    """
    atom_site_dict = file.get_category("atom_site", data_block)
    return len(_get_model_starts(atom_site_dict["pdbx_PDB_model_num"]))


def get_structure(pdbx_file, model=None, data_block=None, altloc="first",
                  extra_fields=None, use_author_fields=True):
    """
    Create an :class:`AtomArray` or :class:`AtomArrayStack` from the
    ``atom_site`` category in a :class:`PDBxFile`.

    Parameters
    ----------
    pdbx_file : PDBxFile
        The file object.
    model : int, optional
        If this parameter is given, the function will return an
        :class:`AtomArray` from the atoms corresponding to the given
        model number (starting at 1).
        Negative values are used to index models starting from the last
        model insted of the first model.
        If this parameter is omitted, an :class:`AtomArrayStack`
        containing all models will be returned, even if the structure
        contains only one model.
    data_block : str, optional
        The name of the data block. Default is the first
        (and most times only) data block of the file.
    altloc : {'first', 'occupancy', 'all'}
        This parameter defines how *altloc* IDs are handled:
            - ``'first'`` - Use atoms that have the first *altloc* ID
              appearing in a residue.
            - ``'occupancy'`` - Use atoms that have the *altloc* ID
              with the highest occupancy for a residue.
            - ``'all'`` - Use all atoms.
              Note that this leads to duplicate atoms.
              When this option is chosen, the ``altloc_id`` annotation
              array is added to the returned structure.
    extra_fields : list of str, optional
        The strings in the list are entry names, that are
        additionally added as annotation arrays.
        The annotation category name will be the same as the PDBx
        subcategory name.
        The array type is always `str`.
        An exception are the special field identifiers:
        ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and ``'charge'``.
        These will convert the fitting subcategory into an
        annotation array with reasonable type.
    use_author_fields : bool, optional
        Some fields can be read from two alternative sources,
        for example both, ``label_seq_id`` and ``auth_seq_id`` describe
        the ID of the residue.
        While, the ``label_xxx`` fields can be used as official pointers
        to other categories in the :class:`PDBxFile`, the ``auth_xxx``
        fields are set by the author(s) of the structure and are
        consistent with the corresponding values in PDB files.
        If `use_author_fields` is true, the annotation arrays will be
        read from the ``auth_xxx`` fields (if applicable),
        otherwise from the the ``label_xxx`` fields.

    Returns
    -------
    array : AtomArray or AtomArrayStack
        The return type depends on the `model` parameter.

    Examples
    --------

    >>> import os.path
    >>> file = PDBxFile.read(os.path.join(path_to_structures, "1l2y.cif"))
    >>> arr = get_structure(file, model=1)
    >>> print(len(arr))
    304

    """
    altloc = [] if altloc is None else altloc
    extra_fields = [] if extra_fields is None else extra_fields

    atom_site_dict = pdbx_file.get_category("atom_site", data_block)
    models = atom_site_dict["pdbx_PDB_model_num"]
    model_starts = _get_model_starts(models)
    model_count = len(model_starts)
    atom_count = len(models)

    if model is None:
        # For a stack, the annotations are derived from the first model
        model_dict = _get_model_dict(atom_site_dict, model_starts, 1)
        # Any field of the category would work here to get the length
        model_length = len(model_dict["group_PDB"])
        stack = AtomArrayStack(model_count, model_length)

        _fill_annotations(stack, model_dict, extra_fields, use_author_fields)

        # Check if each model has the same amount of atoms
        # If not, raise exception
        if model_length * model_count != atom_count:
            raise InvalidFileError(
                "The models in the file have unequal "
                "amount of atoms, give an explicit model "
                "instead"
            )

        stack.coord = np.zeros(
            (model_count, model_length, 3), dtype=np.float32
        )
        stack.coord[:, :, 0] = atom_site_dict["Cartn_x"].reshape(
            (model_count, model_length)
        )
        stack.coord[:, :, 1] = atom_site_dict["Cartn_y"].reshape(
            (model_count, model_length)
        )
        stack.coord[:, :, 2] = atom_site_dict["Cartn_z"].reshape(
            (model_count, model_length)
        )

        stack = _filter_altloc(stack, model_dict, altloc)

        box = _get_box(pdbx_file, data_block)
        if box is not None:
            # Duplicate same box for each model
            stack.box = np.repeat(box[np.newaxis, ...], model_count, axis=0)

        return stack

    else:
        if model == 0:
            raise ValueError("The model index must not be 0")
        # Negative models mean model indexing starting from last model
        model = model_count + model + 1 if model < 0 else model
        if model > model_count:
            raise ValueError(
                f"The file has {model_count} models, "
                f"the given model {model} does not exist"
            )

        model_dict = _get_model_dict(atom_site_dict, model_starts, model)
        # Any field of the category would work here to get the length
        model_length = len(model_dict["group_PDB"])
        array = AtomArray(model_length)

        _fill_annotations(array, model_dict, extra_fields, use_author_fields)

        # Append exclusive stop
        model_starts = np.append(
            model_starts, [len(atom_site_dict["group_PDB"])]
        )
        # Indexing starts at 0, but model number starts at 1
        model_index = model - 1
        start, stop = model_starts[model_index], model_starts[model_index + 1]
        array.coord = np.zeros((model_length, 3), dtype=np.float32)
        array.coord[:, 0] = atom_site_dict["Cartn_x"][start:stop].astype(
            np.float32
        )
        array.coord[:, 1] = atom_site_dict["Cartn_y"][start:stop].astype(
            np.float32
        )
        array.coord[:, 2] = atom_site_dict["Cartn_z"][start:stop].astype(
            np.float32
        )

        array = _filter_altloc(array, model_dict, altloc)

        array.box = _get_box(pdbx_file, data_block)

        return array


def _fill_annotations(array, model_dict, extra_fields, use_author_fields):
    """Fill atom_site annotations in atom array or atom array stack.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        Atom array or stack which will be annotated.
    model_dict : dict(str, ndarray)
        ``atom_site`` dictionary with values for one model.
    extra_fields : list of str
        Entry names, that are additionally added as annotation arrays.
    use_author_fields : bool
        Define if alternate fields prefixed with ``auth_`` should be used
        instead of ``label_``.
    """

    def get_or_fallback_from_dict(input_dict, key, fallback_key,
                                  dict_name="input"):
        """
        Return value related to key in input dict if it exists,
        otherwise try to get the value related to fallback key."""
        if key not in input_dict:
            warnings.warn(
                f"Attribute '{key}' not found within '{dict_name}' category. "
                f"The fallback attribute '{fallback_key}' will be used instead",
                UserWarning
            )
            try:
                return input_dict[fallback_key]
            except KeyError as key_exc:
                raise InvalidFileError(
                    f"Fallback attribute '{fallback_key}' not found in "
                    "'{dict_name}' category"
                ) from key_exc
        return input_dict[key]

    def get_annotation_from_model(
        model_dict,
        annotation_name,
        annotation_fallback=None,
        as_type=None,
        formatter=None,
    ):
        """Get and format annotation array from model dictionary."""
        array = (
            get_or_fallback_from_dict(
                model_dict, annotation_name, annotation_fallback,
                dict_name="atom_site"
            )
            if annotation_fallback is not None
            else model_dict[annotation_name]
        )
        if as_type is not None:
            array = array.astype(as_type)
        return formatter(array) if formatter is not None else array

    prefix, alt_prefix = (
        ("auth", "label") if use_author_fields else ("label", "auth")
    )

    annotation_data = {
        "chain_id": (f"{prefix}_asym_id", f"{alt_prefix}_asym_id", "U4", None),
        "res_id": (
            f"{prefix}_seq_id",
            f"{alt_prefix}_seq_id",
            None,
            lambda annot: np.array(
                [-1 if elt in [".", "?"] else int(elt) for elt in annot]
            ),
        ),
        "ins_code": (
            "pdbx_PDB_ins_code",
            None,
            "U1",
            lambda annot: np.array(
                ["" if elt in [".", "?"] else elt for elt in annot]
            ),
        ),
        "res_name": (f"{prefix}_comp_id", f"{alt_prefix}_comp_id", "U3", None),
        "hetero": ("group_PDB", None, None, lambda annot: annot == "HETATM"),
        "atom_name": (
            f"{prefix}_atom_id",
            f"{alt_prefix}_atom_id",
            "U6",
            None,
        ),
        "element": ("type_symbol", None, "U2", None),
        "atom_id": ("id", None, int, None),
        "b_factor": ("B_iso_or_equiv", None, float, None),
        "occupancy": ("occupancy", None, float, None),
        "charge": (
            "pdbx_formal_charge",
            None,
            None,
            lambda annot: np.array(
                [
                    0 if charge in ["?", "."] else int(charge)
                    for charge in annot
                ],
                dtype=int,
            ),
        ),
    }

    mandatory_annotations = [
        "chain_id",
        "res_id",
        "ins_code",
        "res_name",
        "hetero",
        "atom_name",
        "element",
    ]

    # Iterate over mandatory annotations and given extra_fields
    for annotation_name in mandatory_annotations + extra_fields:
        array.set_annotation(
            annotation_name,
            get_annotation_from_model(
                model_dict, *annotation_data[annotation_name]
            )
            if annotation_name in annotation_data
            else get_annotation_from_model(
                model_dict, annotation_name, as_type=str
            ),
        )


def _filter_altloc(array, model_dict, altloc):
    altloc_ids = model_dict.get("label_alt_id")
    occupancy = model_dict.get("occupancy")

    # Filter altloc IDs and return
    if altloc_ids is None:
        return array
    elif altloc == "occupancy" and occupancy is not None:
        return array[
            ...,
            filter_highest_occupancy_altloc(
                array, altloc_ids, occupancy.astype(float)
            ),
        ]
    # 'first' is also fallback if file has no occupancy information
    elif altloc == "first":
        return array[..., filter_first_altloc(array, altloc_ids)]
    elif altloc == "all":
        array.set_annotation("altloc_id", altloc_ids)
        return array
    else:
        raise ValueError(f"'{altloc}' is not a valid 'altloc' option")


def _get_model_starts(model_array):
    """
    Get the start index for each model in the arrays of the
    ``atom_site`` category.
    """
    models, indices = np.unique(model_array, return_index=True)
    indices.sort()
    return indices


def _get_model_dict(atom_site_dict, model_starts, model):
    """
    Reduce the ``atom_site`` dictionary to the values for the given
    model.
    """
    # Append exclusive stop
    model_starts = np.append(
        model_starts, [len(atom_site_dict["pdbx_PDB_model_num"])]
    )
    model_dict = {}
    # Indexing starts at 0, but model number starts at 1
    model_index = model - 1
    for key in atom_site_dict.keys():
        model_dict[key] = atom_site_dict[key][
            model_starts[model_index] : model_starts[model_index + 1]
        ]
    return model_dict


def _get_box(pdbx_file, data_block):
    if data_block is None:
        cell_dict = pdbx_file.get("cell")
    else:
        cell_dict = pdbx_file.get((data_block, "cell"))
    if cell_dict is None:
        return None
    try:
        len_a, len_b, len_c = [
            float(cell_dict[length])
            for length in ["length_a", "length_b", "length_c"]
        ]
    except ValueError:
        # 'cell_dict' has no proper unit cell values, e.g. '?'
        return None
    alpha, beta, gamma = [
        np.deg2rad(float(cell_dict[angle]))
        for angle in ["angle_alpha", "angle_beta", "angle_gamma"]
    ]
    return vectors_from_unitcell(len_a, len_b, len_c, alpha, beta, gamma)


def set_structure(pdbx_file, array, data_block=None):
    """
    Set the `atom_site` category with an
    :class:`AtomArray` or :class:`AtomArrayStack`.

    This will save the coordinates, the mandatory annotation categories
    and the optional annotation categories
    ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and ``'charge'``.
    If the atom array (stack) contains the annotation ``'atom_id'``,
    these values will be used for atom numbering instead of continuous
    numbering.

    Parameters
    ----------
    pdbx_file : PDBxFile
        The file object.
    array : AtomArray or AtomArrayStack
        The structure to be written. If a stack is given, each array in
        the stack will be in a separate model.
    data_block : str, optional
        The name of the data block. Default is the first
        (and most times only) data block of the file.

    Examples
    --------

    >>> import os.path
    >>> file = PDBxFile()
    >>> set_structure(file, atom_array, data_block="structure")
    >>> file.write(os.path.join(path_to_directory, "structure.cif"))

    """
    # Fill PDBx columns from information
    # in structures' attribute arrays as good as possible
    # Use OrderedDict in order to ensure the usually used column order.
    atom_site_dict = OrderedDict()
    # Save list of annotation categories for checks,
    # if an optional category exists
    annot_categories = array.get_annotation_categories()
    atom_site_dict["group_PDB"] = np.array(
        ["ATOM" if e == False else "HETATM" for e in array.hetero]
    )
    atom_site_dict["type_symbol"] = np.copy(array.element)
    atom_site_dict["label_atom_id"] = np.copy(array.atom_name)
    atom_site_dict["label_alt_id"] = np.full(array.array_length(), ".")
    atom_site_dict["label_comp_id"] = np.copy(array.res_name)
    atom_site_dict["label_asym_id"] = np.copy(array.chain_id)
    atom_site_dict["label_entity_id"] = _determine_entity_id(array.chain_id)
    atom_site_dict["label_seq_id"] = np.array([str(e) for e in array.res_id])
    atom_site_dict["pdbx_PDB_ins_code"] = array.ins_code
    atom_site_dict["auth_seq_id"] = atom_site_dict["label_seq_id"]
    atom_site_dict["auth_comp_id"] = atom_site_dict["label_comp_id"]
    atom_site_dict["auth_asym_id"] = atom_site_dict["label_asym_id"]
    atom_site_dict["auth_atom_id"] = atom_site_dict["label_atom_id"]

    if "atom_id" in annot_categories:
        atom_site_dict["id"] = np.array([str(e) for e in array.atom_id])
    else:
        atom_site_dict["id"] = None
    if "b_factor" in annot_categories:
        atom_site_dict["B_iso_or_equiv"] = np.array(
            [f"{b:.2f}" for b in array.b_factor]
        )
    if "occupancy" in annot_categories:
        atom_site_dict["occupancy"] = np.array(
            [f"{occ:.2f}" for occ in array.occupancy]
        )
    if "charge" in annot_categories:
        atom_site_dict["pdbx_formal_charge"] = np.array(
            [f"{c:+d}" if c != 0 else "?" for c in array.charge]
        )

    # In case of a single model handle each coordinate
    # simply like a flattened array
    if type(array) == AtomArray or (
        type(array) == AtomArrayStack and array.stack_depth() == 1
    ):
        # 'ravel' flattens coord without copy
        # in case of stack with stack_depth = 1
        atom_site_dict["Cartn_x"] = np.array(
            [f"{c:.3f}" for c in np.ravel(array.coord[..., 0])]
        )
        atom_site_dict["Cartn_y"] = np.array(
            [f"{c:.3f}" for c in np.ravel(array.coord[..., 1])]
        )
        atom_site_dict["Cartn_z"] = np.array(
            [f"{c:.3f}" for c in np.ravel(array.coord[..., 2])]
        )
        atom_site_dict["pdbx_PDB_model_num"] = np.full(
            array.array_length(), "1"
        )
    # In case of multiple models repeat annotations
    # and use model specific coordinates
    elif type(array) == AtomArrayStack:
        for key, value in atom_site_dict.items():
            atom_site_dict[key] = np.tile(value, reps=array.stack_depth())
        coord = np.reshape(
            array.coord, (array.stack_depth() * array.array_length(), 3)
        )
        atom_site_dict["Cartn_x"] = np.array([f"{c:.3f}" for c in coord[:, 0]])
        atom_site_dict["Cartn_y"] = np.array([f"{c:.3f}" for c in coord[:, 1]])
        atom_site_dict["Cartn_z"] = np.array([f"{c:.3f}" for c in coord[:, 2]])
        models = np.repeat(
            np.arange(1, array.stack_depth() + 1).astype(str),
            repeats=array.array_length(),
        )
        atom_site_dict["pdbx_PDB_model_num"] = models
    else:
        raise ValueError("Structure must be AtomArray or AtomArrayStack")
    if not "atom_id" in annot_categories:
        # Count from 1
        atom_site_dict["id"] = np.arange(
            1, len(atom_site_dict["group_PDB"]) + 1
        ).astype("U6")
    if data_block is None:
        data_blocks = pdbx_file.get_block_names()
        if len(data_blocks) == 0:
            raise TypeError(
                "No data block is existent in PDBx file, must be specified"
            )
        else:
            data_block = data_blocks[0]
    pdbx_file.set_category("atom_site", atom_site_dict, data_block)

    # Write box into file
    if array.box is not None:
        # PDBx files can only store one box for all models
        # -> Use first box
        if array.box.ndim == 3:
            box = array.box[0]
        else:
            box = array.box
        len_a, len_b, len_c, alpha, beta, gamma = unitcell_from_vectors(box)
        cell_dict = OrderedDict()
        cell_dict["length_a"] = "{:6.3f}".format(len_a)
        cell_dict["length_b"] = "{:6.3f}".format(len_b)
        cell_dict["length_c"] = "{:6.3f}".format(len_c)
        cell_dict["angle_alpha"] = "{:5.3f}".format(np.rad2deg(alpha))
        cell_dict["angle_beta"] = "{:5.3f}".format(np.rad2deg(beta))
        cell_dict["angle_gamma"] = "{:5.3f}".format(np.rad2deg(gamma))
        pdbx_file.set_category("cell", cell_dict, data_block)


def _determine_entity_id(chain_id):
    entity_id = np.zeros(len(chain_id), dtype=int)
    # Dictionary that translates chain_id to entity_id
    id_translation = {}
    id = 1
    for i in range(len(chain_id)):
        try:
            entity_id[i] = id_translation[chain_id[i]]
        except:
            # chain_id is not in dictionary -> new entry
            id_translation[chain_id[i]] = id
            entity_id[i] = id_translation[chain_id[i]]
            id += 1
    return entity_id.astype(str)


def list_assemblies(pdbx_file, data_block=None):
    """
    List the biological assemblies that are available for the structure
    in the given file.

    This function receives the data from the ``pdbx_struct_assembly``
    category in the file.
    Consequently, this category must be present in the file.

    Parameters
    ----------
    pdbx_file : PDBxFile
        The file object.
    data_block : str, optional
        The name of the data block.
        Defaults to the first (and most times only) data block of the
        file.

    Returns
    -------
    assemblies : dict of str -> str
        A dictionary that maps an assembly ID to a description of the
        corresponding assembly.
    """
    assembly_category = pdbx_file.get_category(
        "pdbx_struct_assembly", data_block, expect_looped=True
    )
    if assembly_category is None:
        raise InvalidFileError("File has no 'pdbx_struct_assembly' category")
    return {
        id: details
        for id, details in zip(
            assembly_category["id"], assembly_category["details"]
        )
    }


def get_assembly(pdbx_file, assembly_id=None, model=None, data_block=None,
                 altloc="first", extra_fields=None, use_author_fields=True):
    """
    Build the given biological assembly.

    This function receives the data from the
    ``pdbx_struct_assembly_gen``, ``pdbx_struct_oper_list`` and
    ``atom_site`` categories in the file.
    Consequently, these categories must be present in the file.

    Parameters
    ----------
    pdbx_file : PDBxFile
        The file object.
    assembly_id : str
        The assembly to build
    model : int, optional
        If this parameter is given, the function will return an
        :class:`AtomArray` from the atoms corresponding to the given
        model number (starting at 1).
        Negative values are used to index models starting from the last
        model insted of the first model.
        If this parameter is omitted, an :class:`AtomArrayStack`
        containing all models will be returned, even if the structure
        contains only one model.
    data_block : str, optional
        The name of the data block.
        Defaults to the first (and most times only) data block of the
        file.
    altloc : {'first', 'occupancy', 'all'}
        This parameter defines how *altloc* IDs are handled:
            - ``'first'`` - Use atoms that have the first *altloc* ID
              appearing in a residue.
            - ``'occupancy'`` - Use atoms that have the *altloc* ID
              with the highest occupancy for a residue.
            - ``'all'`` - Use all atoms.
              Note that this leads to duplicate atoms.
              When this option is chosen, the ``altloc_id`` annotation
              array is added to the returned structure.
    extra_fields : list of str, optional
        The strings in the list are entry names, that are
        additionally added as annotation arrays.
        The annotation category name will be the same as the PDBx
        subcategory name.
        The array type is always `str`.
        An exception are the special field identifiers:
        ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and ``'charge'``.
        These will convert the fitting subcategory into an
        annotation array with reasonable type.
    use_author_fields : bool, optional
        Some fields can be read from two alternative sources,
        for example both, ``label_seq_id`` and ``auth_seq_id`` describe
        the ID of the residue.
        While, the ``label_xxx`` fields can be used as official pointers
        to other categories in the :class:`PDBxFile`, the ``auth_xxx``
        fields are set by the author(s) of the structure and are
        consistent with the corresponding values in PDB files.
        If `use_author_fields` is true, the annotation arrays will be
        read from the ``auth_xxx`` fields (if applicable),
        otherwise from the the ``label_xxx`` fields.

    Returns
    -------
    assembly : AtomArray or AtomArrayStack
        The assembly. The return type depends on the `model` parameter.
    """
    assembly_gen_category = pdbx_file.get_category(
        "pdbx_struct_assembly_gen", data_block, expect_looped=True
    )
    if assembly_gen_category is None:
        raise InvalidFileError(
            "File has no 'pdbx_struct_assembly_gen' category"
        )

    struct_oper_category = pdbx_file.get_category(
        "pdbx_struct_oper_list", data_block, expect_looped=True
    )
    if struct_oper_category is None:
        raise InvalidFileError("File has no 'pdbx_struct_oper_list' category")

    if assembly_id is None:
        assembly_id = assembly_gen_category["assembly_id"][0]
    elif assembly_id not in assembly_gen_category["assembly_id"]:
        raise KeyError(f"File has no Assembly ID '{assembly_id}'")

    transformations = _get_transformations(struct_oper_category)

    ### Get transformations to apply and the affected asym IDs
    # Find the operation expression for given assembly ID
    # We already asserted that the ID is actually present
    for id, op_expr, asym_id_expr in zip(
        assembly_gen_category["assembly_id"],
        assembly_gen_category["oper_expression"],
        assembly_gen_category["asym_id_list"],
    ):
        if id == assembly_id:
            operations = _parse_operation_expression(op_expr)
            asym_ids = asym_id_expr.split(",")
            break

    ### Get structure containing the affected asym IDs
    # Include 'label_asym_id' as annotation array
    # for correct asym ID filtering
    extra_fields = [] if extra_fields is None else extra_fields
    if "label_asym_id" in extra_fields:
        extra_fields_and_asym = extra_fields
    else:
        extra_fields_and_asym = extra_fields + ["label_asym_id"]
    structure = get_structure(
        pdbx_file,
        model,
        data_block,
        altloc,
        extra_fields_and_asym,
        use_author_fields,
    )
    # Filter asym IDs
    structure = structure[..., np.isin(structure.label_asym_id, asym_ids)]
    # Remove 'label_asym_id', if it was not included in the original
    # user-supplied 'extra_fields'
    if "label_asym_id" not in extra_fields:
        structure.del_annotation("label_asym_id")

    ### Prepare coordinates
    if model is None:
        # Coordinates for AtomArrayStack
        assembly_coord = np.zeros(
            (
                len(operations),
                structure.stack_depth(),
                structure.array_length(),
                3,
            )
        )
    else:
        # Coordinates for AtomArray
        assembly_coord = np.zeros(
            (len(operations), structure.array_length(), 3)
        )

    ### Apply corresponding transformation for each copy in the assembly
    for i, operation in enumerate(operations):
        coord = structure.coord
        # Execute for each transformation step
        # in the operation expression
        for op_step in operation:
            rotation_matrix, translation_vector = transformations[op_step]
            # Rotate
            coord = matrix_rotate(coord, rotation_matrix)
            # Translate
            coord += translation_vector
        assembly_coord[i] = coord

    return repeat(structure, assembly_coord)


def _get_transformations(struct_oper):
    transformation_dict = {}
    for index, id in enumerate(struct_oper["id"]):
        rotation_matrix = np.array(
            [
                [
                    float(struct_oper[f"matrix[{i}][{j}]"][index])
                    for j in (1, 2, 3)
                ]
                for i in (1, 2, 3)
            ]
        )
        translation_vector = np.array(
            [float(struct_oper[f"vector[{i}]"][index]) for i in (1, 2, 3)]
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def _parse_operation_expression(expression):
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for expr in expressions_per_step:
        if "-" in expr:
            # Range of operation IDs, they must be integers
            first, last = expr.split("-")
            operations.append(
                [str(id) for id in range(int(first), int(last) + 1)]
            )
        elif "," in expr:
            # List of operation IDs
            operations.append(expr.split(","))
        else:
            # Single operation ID
            operations.append([expr])

    # Cartesian product of operations
    return list(itertools.product(*operations))


def _convert_string_to_sequence(string, stype):
    # Convert strings to ProteinSequence-Object if stype is
    # contained in _proteinseq_type_list or to NucleotideSequence-
    # Object if stype is contained in _nucleotideseq_type_list
    if stype in _proteinseq_type_list:
        return ProteinSequence(string)
    elif stype in _nucleotideseq_type_list:
        string = string.replace("U", "T")
        return NucleotideSequence(string)
    elif stype in _other_type_list:
        return None
    else:
        raise InvalidFileError(
            "mmCIF _entity_poly.type unsupported" " type: " + stype
        )



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Utility functions for in internal use in `Bio.Structure` package
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["vector_dot", "norm_vector", "distance", "matrix_rotate"]

import numpy as np
from .atoms import Atom, array


def vector_dot(v1,v2):
    """
    Calculate vector dot product of two vectors.
    
    Parameters
    ----------
    v1,v2 : ndarray
        The arrays to calculate the product from.
        The vectors are represented by the last axis.
    
    Returns
    -------
    product : float or ndarray
        Scalar product over the last dimension of the arrays.
    """
    return (v1*v2).sum(axis=-1)


def norm_vector(v):
    """
    Normalise a vector.
    
    Parameters
    ----------
    v : ndarray
        The array containg the vector(s).
        The vectors are represented by the last axis.
    """
    factor = np.linalg.norm(v, axis=-1)
    if isinstance(factor, np.ndarray):
        v /= factor[..., np.newaxis]
    else:
        v /= factor
        

def distance(v1,v2):
    """
    Calculate the distance between two position vectors.
    
    Parameters
    ----------
    v1,v2 : ndarray
        The arrays to calculate the product from.
        The vectors are represented by the last axis.
    
    Returns
    -------
    product : float or ndarray
        Vector distance over the last dimension of the array.
    """
    dif = v1 - v2
    return np.sqrt((dif*dif).sum(axis=-1))


def matrix_rotate(v, matrix):
    """
    Perform a rotation using a rotation matrix.

    Parameters
    ----------
    v : ndarray
        The coordinates to rotate.
    matrix : ndarray
        The rotation matrix.
    
    Returns
    -------
    rotated : ndarray
        The rotated coordinates.
    """
    # For proper rotation reshape into a maximum of 2 dimensions
    orig_ndim = v.ndim
    if orig_ndim > 2:
        orig_shape = v.shape
        v = v.reshape(-1, 3)
    # Apply rotation
    v = np.dot(matrix, v.T).T
    # Reshape back into original shape
    if orig_ndim > 2:
        v = v.reshape(*orig_shape)
    return v




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing an :class:`AtomArray` or
:class:`AtomArrayStack` using the popular PDB format.
Since this format has some limitations, it will be completely replaced
someday by the modern PDBx/mmCIF format.
Therefore this subpackage should only be used, if usage of the
PDBx/mmCIF format is not suitable (e.g. when interfacing
other software).
In all other cases, usage of the ``pdbx`` subpackage is encouraged.
"""

__name__ = "biotite.structure.io.pdb"
__author__ = "Patrick Kunzmann"

from .file import *
from .convert import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading/writing information
(especially sequence features) from/to files in the *GenBank*
and *GenPept* format.
"""

__name__ = "biotite.sequence.io.genbank"
__author__ = "Patrick Kunzmann"

from .file import *
from .annotation import *
from .sequence import *
from .metadata import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.viennarna"
__author__ = "Tom David Müller"
__all__ = ["RNAplotApp"]

import numpy as np
from tempfile import NamedTemporaryFile
from os import remove
from enum import IntEnum
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...structure.dotbracket import dot_bracket as dot_bracket_

class RNAplotApp(LocalApp):
    """
    Get coordinates for a 2D representation of any unknotted RNA
    structure using *ViennaRNA's* *RNAplot*.

    The structure has to be provided either in dot bracket notation or
    as a ``ndarray`` of base pairs and the total sequence length.

    Internally this creates a :class:`Popen` instance, which handles
    the execution.

    Parameters
    ----------
    dot_bracket : str, optional (default: None)
        The structure in dot bracket notation.
    base_pairs : ndarray, shape=(n,2), optional (default: None)
        Each row corresponds to the positions of the bases in the
        strand. This parameter is mutually exclusive to ``dot_bracket``.
    length : int, optional (default: None)
        The number of bases in the strand. This parameter is required if
        ``base_pairs`` is given.
    layout_type : RNAplotApp.Layout, optional (default: RNAplotApp.Layout.NAVIEW)
        The layout type according to the *RNAplot* documentation.
    bin_path : str, optional
        Path of the *RNAplot* binary.

    Examples
    --------

    >>> app = RNAplotApp('((..))')
    >>> app.start()
    >>> app.join()
    >>> print(app.get_coordinates())
    [[ -92.50   92.50]
     [ -92.50   77.50]
     [ -90.31   58.24]
     [-109.69   58.24]
     [-107.50   77.50]
     [-107.50   92.50]]
    """

    class Layout(IntEnum):
        """
        This enum type represents the layout type of the plot according
        to the official *RNAplot* orientation.
        """
        RADIAL = 0,
        NAVIEW = 1,
        CIRCULAR = 2,
        RNATURTLE = 3,
        RNAPUZZLER = 4

    def __init__(self, dot_bracket=None, base_pairs=None, length=None,
                 layout_type=Layout.NAVIEW, bin_path="RNAplot"):
        super().__init__(bin_path)

        if dot_bracket is not None:
            self._dot_bracket = dot_bracket
        elif (base_pairs is not None) and (length is not None):
            self._dot_bracket = dot_bracket_(
                base_pairs, length, max_pseudoknot_order = 0
            )[0]
        else:
            raise ValueError(
                "Structure has to be provided in either dot bracket notation "
                "or as base pairs and total sequence length"
            )

        # Get the value of the enum type
        self._layout_type = str(int(layout_type))
        self._in_file  = NamedTemporaryFile("w", suffix=".fold",  delete=False)

    def run(self):
        self._in_file.write("N"*len(self._dot_bracket) + "\n")
        self._in_file.write(self._dot_bracket)
        self._in_file.flush()
        self.set_arguments(
            ["-i", self._in_file.name, "-o", "xrna", "-t", self._layout_type]
        )
        super().run()

    def evaluate(self):
        super().evaluate()
        self._coordinates = np.loadtxt("rna.ss", usecols=(2, 3))

    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_file)
        remove("rna.ss")

    @requires_state(AppState.CREATED)
    def set_layout_type(self, layout_type):
        """
        Adjust the layout type for the plot according to the *RNAplot*
        documentation.

        Parameters
        ----------
        type : RNAplotApp.Layout
            The layout type.
        """
        self._layout_type = str(layout_type)

    @requires_state(AppState.JOINED)
    def get_coordinates(self):
        """
        Get coordinates for a 2D representation of the input structure.

        Returns
        -------
        coordinates : ndarray, shape=(n,2)
            The 2D coordinates. Each row represents the *x* and *y*
            coordinates for a total sequence length of *n*.

        Examples
        --------

        >>> app = RNAplotApp('((..))')
        >>> app.start()
        >>> app.join()
        >>> print(app.get_coordinates())
        [[ -92.50   92.50]
         [ -92.50   77.50]
         [ -90.31   58.24]
         [-109.69   58.24]
         [-107.50   77.50]
         [-107.50   92.50]]
        """
        return self._coordinates

    @staticmethod
    def compute_coordinates(
        dot_bracket=None, base_pairs=None, length=None,
        layout_type=Layout.NAVIEW, bin_path="RNAplot"
    ):
        """
        Get coordinates for a 2D representation of any unknotted RNA
        structure using *ViennaRNA's* *RNAplot*.

        The structure has to be provided either in dot bracket notation
        or as a ``ndarray`` of base pairs and the total sequence length.

        This is a convenience function, that wraps the
        :class:`RNAplotApp` execution.

        Parameters
        ----------
        dot_bracket : str, optional (default: None)
            The structure in dot bracket notation.
        base_pairs : ndarray, shape=(n,2), optional (default: None)
            Each row corresponds to the positions of the bases in the
            strand. This parameter is mutually exclusive to
            ``dot_bracket``.
        length : int, optional (default: None)
            The number of bases in the strand. This parameter is
            required if ``base_pairs`` is given.
        bin_path : str, optional
            Path of the *RNAplot* binary.

        Returns
        -------
        coordinates : ndarray, shape=(n,2)
            The 2D coordinates. Each row represents the *x* and *y*
            coordinates for a total sequence length of *n*.
        """
        app = RNAplotApp(dot_bracket=dot_bracket, base_pairs=base_pairs,
                         length=length, layout_type=layout_type,
                         bin_path=bin_path)
        app.start()
        app.join()
        return app.get_coordinates()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.sequence as seq
import pytest


@pytest.mark.parametrize("table_id",
    [1,2,3,4,5,6,9,10,11,12,13,14,16,21,22,23,24,25,26,27,28,29,30,31])
def test_table_load(table_id):
    table = seq.CodonTable.load(table_id)


def test_table_indexing():
    table = seq.CodonTable.load("Standard")
    assert table["ATG"] == "M"
    for codon in table["Y"]:
        assert codon in ("TAT", "TAC")
    assert table[(0, 0, 0)] == 8
    for codon in table[8]:
        assert codon in ((0, 0, 0), (0, 0, 2))


def test_incomplete_input_dict():
    """
    Ensure that an exception is raised when the input codon dictionary
    misses an entry.
    """
    with pytest.raises(ValueError, match=".* dictionary does not contain .*"):
        codon_dict = seq.CodonTable.default_table().codon_dict()
        del codon_dict["CTG"]
        seq.CodonTable(codon_dict, starts=["ATG"])


def test_codon_dict():
    """
    Assert that the 'codon_dict()' method returns a correct dictionary.
    """
    default_table = seq.CodonTable.default_table()
    ref_codon_dict = default_table.codon_dict()
    test_codon_dict = seq.CodonTable(
        ref_codon_dict, default_table.start_codons()
    ).codon_dict()
    assert ref_codon_dict == test_codon_dict



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.uniprot"
__author__ = "Maximilian Greil"
__all__ = ["assert_valid_response"]

from ..error import RequestError


# Taken from https://www.uniprot.org/help/api_retrieve_entries
def assert_valid_response(response_status_code):
    """
    Checks whether the response is valid.

    Parameters
    ----------
    response_status_code: int
        Status code of request.get.
    """
    if response_status_code == 400:
        raise RequestError("Bad request. There is a problem with your input.")
    elif response_status_code == 404:
        raise RequestError("Not found. The resource you requested doesn't exist.")
    elif response_status_code == 410:
        raise RequestError("Gone. The resource you requested was removed.")
    elif response_status_code == 500:
        raise RequestError(
            "Internal server error. Most likely a temporary problem, but if the problem persists please contact UniProt team.")
    elif response_status_code == 503:
        raise RequestError("Service not available. The server is being updated, try again later.")



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["vdw_radius_protor", "vdw_radius_single"]

from .bonds import bonds_in_residue


# Contains tuples for the different ProtOr groups:
# Tuple contains: element, valency, H count
_PROTOR_RADII = {
    ("C",  3, 0) : 1.61,
    ("C",  3, 1) : 1.76,
    ("C",  4, 1) : 1.88,
    ("C",  4, 2) : 1.88,
    ("C",  4, 3) : 1.88,
    ("N",  3, 0) : 1.64,
    ("N",  3, 1) : 1.64,
    ("N",  3, 2) : 1.64,
    ("N",  4, 3) : 1.64,
    ("O",  1, 0) : 1.42,
    ("O",  2, 1) : 1.46,
    ("S",  1, 0) : 1.77,
    ("S",  2, 0) : 1.77, # Not official, added for completeness (MET)
    ("S",  2, 1) : 1.77,
    ("F",  1, 0) : 1.47, # Taken from _SINGLE_RADII
    ("CL", 1, 0) : 1.75, # Taken from _SINGLE_RADII
    ("BR", 1, 0) : 1.85, # Taken from _SINGLE_RADII
    ("I",  1, 0) : 1.98, # Taken from _SINGLE_RADII
}

_SINGLE_RADII = {
    "H":  1.20,
    "HE": 1.40,
    
    "C":  1.70,
    "N":  1.55,
    "O":  1.52,
    "F":  1.47,
    "NE": 1.54,
    
    "SI": 2.10,
    "P":  1.80,
    "S":  1.80,
    "CL": 1.75,
    "AR": 1.88,
    
    "AS": 1.85,
    "SE": 1.90,
    "BR": 1.85,
    "KR": 2.02,
    
    "TE": 2.06,
    "I":  1.98,
    "XE": 2.16,
}

# A dictionary that caches radii for each residue
_protor_radii = {}


def vdw_radius_protor(res_name, atom_name):
    """
    Estimate the Van-der-Waals radius of an non-hydrogen atom,
    that includes the radius added by potential bonded hydrogen atoms.
    The respective radii are taken from the ProtOr dataset.
    :footcite:`Tsai1999`

    This is especially useful for macromolecular structures where no
    hydrogen atoms are resolved, e.g. crystal structures.
    The valency of the non-hydrogen atom and the amount of normally
    bonded hydrogen atoms is taken from the chemical compound dictionary
    dataset.

    Parameters
    ----------
    res_name : str
        The up to 3-letter residue name the non-hydrogen atom belongs
        to.
    atom_name : str
        The name of the non-hydrogen atom.
    
    Returns
    -------
    The Van-der-Waals radius of the given atom.
    If the radius cannot be estimated for the atom, `None` is returned.

    See also
    --------
    vdw_radius_single
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------

    >>> print(vdw_radius_protor("GLY", "CA"))
    1.88
    """
    res_name = res_name.upper()
    if atom_name[0] == "H":
        raise ValueError(
            f"Calculating the ProtOr radius for the hydrogen atom "
            f"'{atom_name}' is not meaningful"
        )
    if res_name in _protor_radii:
        # Use cached radii for the residue, if already calculated
        if atom_name not in _protor_radii[res_name]:
            raise KeyError(
                f"Residue '{res_name}' does not contain an atom named "
                f"'{atom_name}'"
            )
        return _protor_radii[res_name].get(atom_name)
    else:
        # Otherwise calculate radii for the given residue and cache
        _protor_radii[res_name] = _calculate_protor_radii(res_name)
        # Recursive call, but this time the radii for the given residue
        # are cached
        return vdw_radius_protor(res_name, atom_name)

def _calculate_protor_radii(res_name):
    """
    Calculate the ProtOr VdW radii for all atoms (atom names) in
    a residue.
    """
    bonds = bonds_in_residue(res_name)
    # Maps atom names to a ProtOr group
    # -> tuple(element, valency, H count)
    # Based on the group the radius is chosen from _PROTOR_RADII
    groups = {}
    for atom1, atom2 in bonds:
        # Process each bond two times:
        # One time the first atom is the one to get valency and H count
        # for and the other time vice versa
        for main_atom, bound_atom in ((atom1, atom2), (atom2, atom1)):
            element = main_atom[0]
            # Calculating ProtOr radii for hydrogens in not meaningful
            if element == "H":
                continue
            # Only for these elements ProtOr groups exist
            # Calculation of group for all other elements would be
            # pointless
            if element not in ["C", "N", "O", "S"]:
                # Empty tuple to indicate nonexistent entry
                groups[main_atom] = ()
                continue
            # Update existing entry if already existing
            group = groups.get(main_atom, [element, 0, 0])
            # Increase valency by one, since the bond entry exists
            group[1] += 1
            # If the atom is bonded to hydrogen, increase H count
            if bound_atom[0] == "H":
                group[2] += 1
            groups[main_atom] = group
    # Get radii based on ProtOr groups
    radii = {atom : _PROTOR_RADII.get(tuple(group))
             for atom, group in groups.items()}
    return radii


def vdw_radius_single(element):
    """
    Get the Van-der-Waals radius of an atom from the given element.
    :footcite:`Bondi1964`

    Parameters
    ----------
    element : str
        The chemical element of the atoms.
    
    Returns
    -------
    The Van-der-Waals radius of the atom.
    If the radius is unknown for the element, `None` is returned.
    
    See also
    --------
    vdw_radius_protor
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------

    >>> print(vdw_radius_single("C"))
    1.7
    """
    return _SINGLE_RADII.get(element.upper())


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Some convenience functions for consistency with other ``structure.io``
subpackages.
"""

__name__ = "biotite.structure.io.pdb"
__author__ = "Patrick Kunzmann"
__all__ = ["get_model_count", "get_structure", "set_structure"]


def get_model_count(pdb_file):
    """
    Get the number of models contained in a :class:`PDBFile`.

    Parameters
    ----------
    pdb_file : PDBFile
        The file object.

    Returns
    -------
    model_count : int
        The number of models.
    """
    return pdb_file.get_model_count()


def get_structure(pdb_file, model=None, altloc="first", extra_fields=[],
                  include_bonds=False):
    """
    Create an :class:`AtomArray` or :class:`AtomArrayStack` from a
    :class:`PDBFile`.

    This function is a thin wrapper around the :class:`PDBFile` method
    :func:`get_structure()` for the sake of consistency with other
    ``structure.io`` subpackages.
    
    Parameters
    ----------
    pdb_file : PDBFile
        The file object.
    model : int, optional
        If this parameter is given, the function will return an
        :class:`AtomArray` from the atoms corresponding to the given
        model number (starting at 1).
        Negative values are used to index models starting from the last
        model insted of the first model.
        If this parameter is omitted, an :class:`AtomArrayStack`
        containing all models will be returned, even if the structure
        contains only one model.
    altloc : {'first', 'occupancy', 'all'}
        This parameter defines how *altloc* IDs are handled:
            - ``'first'`` - Use atoms that have the first *altloc* ID
              appearing in a residue.
            - ``'occupancy'`` - Use atoms that have the *altloc* ID
              with the highest occupancy for a residue.
            - ``'all'`` - Use all atoms.
              Note that this leads to duplicate atoms.
              When this option is chosen, the ``altloc_id`` annotation
              array is added to the returned structure.
    extra_fields : list of str, optional
        The strings in the list are optional annotation categories
        that should be stored in the output array or stack.
        These are valid values:
        ``'atom_id'``, ``'b_factor'``, ``'occupancy'`` and ``'charge'``.
    include_bonds : bool, optional
        If set to true, a :class:`BondList` will be created for the
        resulting :class:`AtomArray` containing the bond information
        from the file.
        All bonds have :attr:`BondType.ANY`, since the PDB format
        does not support bond orders.
        
    Returns
    -------
    array : AtomArray or AtomArrayStack
        The return type depends on the `model` parameter.
    
    """
    return pdb_file.get_structure(model, altloc, extra_fields, include_bonds)


def set_structure(pdb_file, array, hybrid36=False):
    """
    write an :class:`AtomArray` or :class:`AtomArrayStack` into a
    :class:`PDBFile`.

    This function is a thin wrapper around the :class:`PDBFile` method
    :func:`set_structure()` for the sake of consistency with other
    ``structure.io`` subpackages.
    
    This will save the coordinates, the mandatory annotation categories
    and the optional annotation categories
    'atom_id', 'b_factor', 'occupancy' and 'charge'.
    
    Parameters
    ----------
    pdb_file : PDBFile
        The file object.
    array : AtomArray or AtomArrayStack
        The structure to be written. If a stack is given, each array in
        the stack will be in a separate model.
    hybrid36: boolean, optional
        Defines wether the file should be written in hybrid-36 format.

    Notes
    -----
    If `array` has an associated :class:`BondList`, ``CONECT``
    records are also written for all non-water hetero residues
    and all inter-residue connections.
    """
    pdb_file.set_structure(array, hybrid36)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for calculation of mechanical or
mass-related properties of a molecular structure.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["mass_center", "gyration_radius"]

import numpy as np
from .atoms import Atom, AtomArray, AtomArrayStack, coord
from .util import vector_dot, norm_vector
from .error import BadStructureError
from .geometry import distance
from .info.masses import mass


def gyration_radius(array, masses=None):
    """
    Compute the radius/radii of gyration of an atom array or stack.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array or stack to calculate the radius/radii of gyration
        for.
    masses : ndarray, optional
        The masses to use for each atom in the input `array`.
        Must have the same length as `array`. By default, the standard
        atomic mass for each element is taken.

    
    Returns
    -------
    masses : float or ndarray, dtype=float
        If `array` is an :class:`AtomArray`, the radius of gyration is
        returned as single float.
        If `array` is an :class:`AtomArrayStack`, a :class:`ndarray`
        containing the radii of gyration for every model is returned.
    """
    if masses is None:
        masses = np.array([mass(element) for element in array.element])
    center = mass_center(array, masses)
    radii = distance(array, center[..., np.newaxis, :])
    inertia_moment = np.sum(masses * radii*radii, axis=-1)
    return np.sqrt(inertia_moment / np.sum(masses))

def mass_center(array, masses=None):
    """
    Calculate the center(s) of mass of an atom array or stack.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array or stack to calculate the center(s) of mass for.
    masses : ndarray, optional
        The masses to use for each atom in the input `array`.
        Must have the same length as `array`. By default, the standard
        atomic mass for each element is taken.
    
    Returns
    -------
    radius : ndarray, ndarray, dtype=float
        Array containing the the coordinates of the center of mass.
        If `array` is an :class:`AtomArray`, this will be an length 3
        :class:`ndarray`; if it is an :class:`AtomArrayStack` with *n* models,
        a (*n x 3*) :class:`ndarray` is returned.
    """
    if masses is None:
        masses = np.array([mass(element) for element in array.element])
    return np.sum(masses[:,np.newaxis] * array.coord, axis=-2) / np.sum(masses)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import numpy as np
import pytest
import biotite
import biotite.sequence.phylo as phylo
from ..util import data_dir


@pytest.fixture
def distances():
    # Distances are based on the example
    # "Dendrogram of the BLOSUM62 matrix"
    # with the small modification M[i,j] += i+j
    # to reduce ambiguity in the tree construction.
    return np.loadtxt(join(data_dir("sequence"), "distances.txt"), dtype=int)


@pytest.fixture
def upgma_newick():
    # Newick notation of the tree created from 'distances.txt',
    # created via DendroUPGMA
    with open(join(data_dir("sequence"), "newick_upgma.txt"), "r") as file:
        newick = file.read().strip()
    return newick


@pytest.fixture
def tree(distances):
    return phylo.upgma(distances)


def test_upgma(tree, upgma_newick):
    """
    Compare the results of `upgma()` with DendroUPGMA.
    """
    ref_tree = phylo.Tree.from_newick(upgma_newick)
    # Cannot apply direct tree equality assertion because the distance
    # might not be exactly equal due to floating point rounding errors
    for i in range(len(tree)):
        for j in range(len(tree)):
            # Check for equal distances and equal topologies
            assert                   tree.get_distance(i,j) \
                == pytest.approx(ref_tree.get_distance(i,j), abs=1e-3)
            assert     tree.get_distance(i,j, topological=True) \
                == ref_tree.get_distance(i,j, topological=True)


def test_neighbor_joining():
    """
    Compare the results of `neighbor_join()` with a known tree.
    """
    dist = np.array([
        [ 0,  5,  4,  7,  6,  8],
        [ 5,  0,  7, 10,  9, 11],
        [ 4,  7,  0,  7,  6,  8],
        [ 7, 10,  7,  0,  5,  9],
        [ 6,  9,  6,  5,  0,  8],
        [ 8, 11,  8,  9,  8,  0],
    ])

    ref_tree = phylo.Tree(phylo.TreeNode(
        [
            phylo.TreeNode(
                [
                    phylo.TreeNode(
                        [
                            phylo.TreeNode(index=0),
                            phylo.TreeNode(index=1),
                        ],
                        [1,4]
                    ),
                    phylo.TreeNode(index=2),
                ],
                [1, 2]
            ),
            phylo.TreeNode(
                [
                    phylo.TreeNode(index=3),
                    phylo.TreeNode(index=4),
                ],
                [3,2]
            ),
            phylo.TreeNode(index=5),
        ],
        [1,1,5]
    ))

    test_tree = phylo.neighbor_joining(dist)

    assert test_tree == ref_tree


def test_node_distance(tree):
    """
    Test whether the `distance_to()` and `lowest_common_ancestor()` work
    correctly.
    """
    # Tree is created via UPGMA
    # -> The distances to root should be equal for all leaf nodes
    dist = tree.root.distance_to(tree.leaves[0])
    for leaf in tree.leaves:
        assert leaf.distance_to(tree.root) == dist
    # Example topological distances
    assert tree.get_distance(0, 19, True) == 9
    assert tree.get_distance(4,  2, True) == 10
    
    # All pairwise leaf node distances should be sufficient
    # to reconstruct the same tree via UPGMA
    ref_dist_mat = np.zeros((len(tree), len(tree)))
    for i in range(len(tree)):
        for j in range(len(tree)):
            ref_dist_mat[i,j] = tree.get_distance(i,j)
    assert np.allclose(ref_dist_mat, ref_dist_mat.T)
    new_tree = phylo.upgma(ref_dist_mat)
    test_dist_mat = np.zeros((len(tree), len(tree)))
    for i in range(len(tree)):
        for j in range(len(tree)):
            test_dist_mat[i,j] = new_tree.get_distance(i,j)
    assert np.allclose(test_dist_mat, ref_dist_mat)


def test_leaf_list(tree):
    for i, leaf in enumerate(tree.leaves):
        assert i == leaf.index


def test_distances(tree):
    # Tree is created via UPGMA
    # -> The distances to root should be equal for all leaf nodes
    dist = tree.root.distance_to(tree.leaves[0])
    for leaf in tree.leaves:
        assert leaf.distance_to(tree.root) == dist
    # Example topological distances
    assert tree.get_distance(0, 19, True) == 9
    assert tree.get_distance(4,  2, True) == 10


def test_get_leaves(tree):
    # Manual example cases
    node = tree.leaves[6]
    assert set(tree.leaves[6].parent.get_indices()) == set(
        [6,11,2,3,13,8,14,5,0,15,16]
    )
    assert set(tree.leaves[10].get_indices()) == set([10])
    assert tree.root.get_leaf_count() == 20

 
def test_copy(tree):
    assert tree is not tree.copy()
    assert tree == tree.copy()


def test_immutability():
    node = phylo.TreeNode(index=0)
    # Attributes are not writable
    with pytest.raises(AttributeError):
        node.children = None
    with pytest.raises(AttributeError):
        node.parent = None
    with pytest.raises(AttributeError):
        node.index = None
    # A root node cannot be child
    node1 = phylo.TreeNode(index=0)
    node2 = phylo.TreeNode(index=1)
    node1.as_root()
    with pytest.raises(phylo.TreeError):
        phylo.TreeNode([node1, node2], [0, 0])
    # A child node cannot be root
    node1 = phylo.TreeNode(index=0)
    node2 = phylo.TreeNode(index=1)
    phylo.TreeNode([node1, node2], [0, 0])
    with pytest.raises(phylo.TreeError):
        node1.as_root()
    # A node cannot be child of a two nodes
    node1 = phylo.TreeNode(index=0)
    node2 = phylo.TreeNode(index=1)
    phylo.TreeNode([node1, node2], [0, 0])
    with pytest.raises(phylo.TreeError):
        phylo.TreeNode([node1, node2], [0, 0])
    # Tree cannot be constructed from child nodes
    node1 = phylo.TreeNode(index=0)
    node2 = phylo.TreeNode(index=0)
    # node1 and node2 have now a parent
    phylo.TreeNode([node1, node2], [0, 0])
    with pytest.raises(phylo.TreeError):
        phylo.Tree(node1)


@pytest.mark.parametrize("newick, labels, error", [
    # Reference index out of range
    ("((1,0),4),2);", None, biotite.InvalidFileError),
    # Empty string
    ("", None, biotite.InvalidFileError),
    # Empty node
    ("();", None, biotite.InvalidFileError),
    # Missing brackets
    ("((0,1,(2,3));", None, biotite.InvalidFileError),
    # A node with three leaves
    ("((0,1),(2,3),(4,5));", None, None),
    # A node with one leaf
    ("((0,1),(2,3),(4));", None, None),
    # Named intermediate nodes
    ("((0,1,3)A,2)B;", None, None),
    # Named intermediate nodes and distances
    ("((0:1.0,1:3.0,3:5.0)A:2.0,2:5.0)B;", None, None),
    # Nodes with labels
    ("((((A,B),(C,D)),E),F);", ["A","B","C","D","E","F"], None),
    # Nodes with labels and distances
    ("((((A:1,B:2),(C:3,D:4)),E:5),F:6);", ["A","B","C","D","E","F"], None),
    # Newick with spaces
    (" ( 0 : 1.0 , 1 : 3.0 ) A ; ", None, None),
])
def test_newick_simple(newick, labels, error):
    # Read, write and read again a Newick notation and expect
    # the same reult from both reads
    if error is None:
        tree1 = phylo.Tree.from_newick(newick, labels)
        newick = tree1.to_newick(labels, include_distance=True)
        tree2 = phylo.Tree.from_newick(newick, labels)
        assert tree1 == tree2
    else:
         with pytest.raises(error):
             tree1 = phylo.Tree.from_newick(newick, labels)


@pytest.mark.parametrize("use_labels", [False, True])
def test_newick_complex(upgma_newick, use_labels):
    # Same as above with more complex string
    if use_labels:
        labels = [str(i) for i in range(20)]
    else:
        labels = None
    tree1 = phylo.Tree.from_newick(upgma_newick, labels)
    newick = tree1.to_newick(labels, include_distance=True)
    tree2 = phylo.Tree.from_newick(newick, labels)
    assert tree1 == tree2


@pytest.mark.parametrize("newick_in, exp_newick_out", [
    ("(0:1.0, 1:2.0);",                     "(0:1.0,1:2.0):0.0;"             ),
    ("(0:1.0, 1:2.0, 2:3.0);",              "((0:1.0,1:2.0):0.0,2:3.0):0.0;" ),
    ("(((0:1.0, 1:2.0):10.0):5.0, 2:8.0);", "((0:1.0,1:2.0):15.0,2:8.0):0.0;"),
    ("((0:1.0, 1:2.0):10.0):5.0;",          "(0:1.0,1:2.0):0.0;"             ),
])
def test_as_binary_cases(newick_in, exp_newick_out):
    """
    Test the `as_binary()` function based on known cases.
    """
    tree = phylo.Tree.from_newick(newick_in)
    bin_tree = phylo.as_binary(tree)
    assert bin_tree.to_newick() == exp_newick_out


def test_as_binary_distances():
    """
    Test the preservation of all pairwise leaf distances after calling
    `as_binary()`.
    """
    # Some random newick
    newick = "((((0:5, 1:1, 2:13, 5:9):4, (4:2, 6:9):7):18), 3:12);"
    tree = phylo.Tree.from_newick(newick)
    ref_dist_mat = np.zeros((len(tree), len(tree)))
    for i in range(len(tree)):
        for j in range(len(tree)):
            ref_dist_mat[i,j] = tree.get_distance(i,j)
    
    bin_tree = phylo.as_binary(tree)
    test_dist_mat = np.zeros((len(tree), len(tree)))
    for i in range(len(tree)):
        for j in range(len(tree)):
            test_dist_mat[i,j] = bin_tree.get_distance(i,j)
    assert np.allclose(test_dist_mat, ref_dist_mat)


def test_equality(tree):
    """
    Assert that equal trees equal each other, and non-equal trees do not
    equal each other.
    """
    assert tree == tree.copy()
    # Order of children is not important
    assert tree == phylo.Tree(phylo.TreeNode(
        [tree.root.children[1].copy(), tree.root.children[0].copy()],
        [tree.root.children[1].distance, tree.root.children[0].distance]
    ))
    # Different distance -> Unequal tree
    assert tree != phylo.Tree(phylo.TreeNode(
        [tree.root.children[0].copy(), tree.root.children[1].copy()],
        [tree.root.children[0].distance, 42]
    ))
    # Additional node -> Unequal tree
    assert tree != phylo.Tree(phylo.TreeNode(
        [
            tree.root.children[0].copy(),
            tree.root.children[1].copy(),
            phylo.TreeNode(index=len(tree))
        ],
        [
            tree.root.children[0].distance,
            tree.root.children[1].distance,
            42
        ]
    ))




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing sequence features in the
*Generic Feature Format 3* (GFF3).

It provides the :class:`GFFFile` class, a low-level line-based
interface to this format, and high-level functions for extracting
:class:`Annotation` objects.

.. note: This package cannot create hierarchical data structures from
   GFF 3 files. This means, that you cannot directly access the the
   parent or child of a feature.
   However, the ``Id`` and ``Name`` attributes are stored in the
   qualifiers of the created :class:`Feature` objects. 
   Hence, it is possible to implement such a data structure from this
   information.
"""

__name__ = "biotite.sequence.io.gff"
__author__ = "Patrick Kunzmann"

from .file import *
from .convert import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.sequence as seq
import biotite.sequence.io as seqio
import biotite.application.blast as blast
import numpy as np
from requests.exceptions import ConnectionError
import pytest
import os.path
from ..util import data_dir, cannot_connect_to


BLAST_URL = "https://blast.ncbi.nlm.nih.gov/Blast.cgi"


# Start of E. coli lacZ ORF (UID: AJ308295)
dna_seq = seq.NucleotideSequence("ATGACCATGATTACGCCAAGCTTTCCGGGGAATTCA")

# Start of E. coli lacZ, translated dna_seq (UID: AJ308295)
prot_seq = seq.ProteinSequence("MTMITPSFPGNS")


@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_blastn():
    app = blast.BlastWebApp("blastn", dna_seq, obey_rules=False)
    app.set_max_expect_value(100)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    assert dna_seq == alignments[0].sequences[0]
    assert dna_seq == alignments[0].sequences[1]

@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_blastx():
    app = blast.BlastWebApp("blastx", dna_seq, obey_rules=False)
    app.set_max_expect_value(100)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    assert prot_seq == alignments[0].sequences[0]
    assert prot_seq == alignments[0].sequences[1]

@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_tblastx():
    app = blast.BlastWebApp("tblastx", dna_seq, obey_rules=False)
    app.set_max_expect_value(100)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    print (alignments[0].sequences[0])
    print (alignments[0].sequences[1])
    rev_prot_seq = dna_seq.reverse().complement().translate(complete=True)
    assert rev_prot_seq == alignments[0].sequences[0]
    assert rev_prot_seq == alignments[0].sequences[1]

@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_blastp():
    app = blast.BlastWebApp("blastp", prot_seq, obey_rules=False)
    app.set_max_expect_value(100)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    assert prot_seq == alignments[0].sequences[0]
    assert prot_seq == alignments[0].sequences[1]

@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_tblastn():
    app = blast.BlastWebApp("tblastn", prot_seq, obey_rules=False)
    app.set_max_expect_value(200)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    assert prot_seq == alignments[0].sequences[0]
    assert prot_seq == alignments[0].sequences[1]

def test_file_input():
    path = os.path.join(data_dir("sequence"), "prot.fasta")
    app = blast.BlastWebApp("blastp", path, obey_rules=False)

def test_invalid_query():
    with pytest.raises(ValueError):
        app = blast.BlastWebApp("blastn", "ABCDEFGHIJKLMNOP", obey_rules=False)
    with pytest.raises(ValueError):
        app = blast.BlastWebApp("blastp", "ABCDEFGHIJKLMNOP", obey_rules=False)
        
@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_no_hit():
    app = blast.BlastWebApp("blastn", "ACTGTACGAAACTCGGCGTA", obey_rules=False)
    app.set_word_size(20)
    app.start()
    app.join(timeout=300)
    alignments = app.get_alignments()
    # BLAST should find original sequence as best hit
    assert len(alignments) == 0

@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_invalid_input():
    app = blast.BlastWebApp("blastn", dna_seq, obey_rules=False)
    # Set some invalid parameters
    app.set_word_size(-20)
    app.set_substitution_matrix("FOOBAR")
    app.start()
    with pytest.raises(ValueError):
        app.join(timeout=300)


@pytest.mark.skipif(
    cannot_connect_to(BLAST_URL),
    reason="NCBI BLAST is not available"
)
def test_hit_with_selenocysteine():
    # Sequence is taken from issue #344
    query = seqio.load_sequence(
        os.path.join(data_dir("sequence"), "selenocysteine.fasta")
    )
    
    # Expect hit containing selenocysteine when searching Swiss-Prot
    blast_app = blast.BlastWebApp("blastp", query, "swissprot")
    blast_app.start()
    # No AlphabetError should be raised here
    blast_app.join()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
from biotite.application.viennarna import RNAplotApp
from ..util import is_not_installed


@pytest.fixture
def sample_app():
    """
    Provide a `RNAplotApp` object, where *RNAplot* has been executed for
    a sample structure.
    """
    app = RNAplotApp('((..))')
    app.start()
    app.join()
    return app


@pytest.mark.skipif(
    is_not_installed("RNAplot"), reason="RNAplot is not installed"
)
def test_get_cooordinates(sample_app):
    assert (
        np.all(
            sample_app.get_coordinates() == np.array([[ -92.5 ,   92.5 ],
                                                      [ -92.5 ,   77.5 ],
                                                      [ -90.31,   58.24],
                                                      [-109.69,   58.24],
                                                      [-107.5 ,   77.5 ],
                                                      [-107.5 ,   92.5 ]])
        )
    )



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
from biotite.application.autodock import VinaApp
from ..util import data_dir, is_not_installed


@pytest.mark.skipif(
    is_not_installed("vina"), reason="Autodock Vina is not installed"
)
@pytest.mark.parametrize("flexible", [False, True])
def test_docking(flexible):
    """
    Test :class:`VinaApp` for the case of docking biotin to
    streptavidin.
    The output binding pose should be very similar to the pose in the
    PDB structure.
    """
    # A structure of a straptavidin-biotin complex
    mmtf_file = mmtf.MMTFFile.read(join(data_dir("application"), "2rtg.mmtf"))
    structure = mmtf.get_structure(
        mmtf_file, model=1, extra_fields=["charge"], include_bonds=True
    )
    structure = structure[structure.chain_id == "B"]
    receptor = structure[struc.filter_amino_acids(structure)]
    ref_ligand = structure[structure.res_name == "BTN"]
    ref_ligand_coord = ref_ligand.coord

    ligand = info.residue("BTN")
    # Remove hydrogen atom that is missing in ref_ligand
    ligand = ligand[ligand.atom_name != "HO2"]

    if flexible:
        # Two residues within the binding pocket: ASN23, SER88
        flexible_mask = np.isin(receptor.res_id, (23, 88))
    else:
        flexible_mask = None
    
    app = VinaApp(
        ligand, receptor, struc.centroid(ref_ligand), [20, 20, 20],
        flexible=flexible_mask
    )
    app.set_seed(0)
    app.start()
    app.join()
    
    test_ligand_coord = app.get_ligand_coord()
    test_receptor_coord = app.get_receptor_coord()
    energies = app.get_energies()
    # One energy value per model
    assert len(test_ligand_coord) == len(energies)
    assert len(test_receptor_coord) == len(energies)

    assert np.all(energies < 0)

    # Select best binding pose
    test_ligand_coord = test_ligand_coord[0]
    not_nan_mask = ~np.isnan(test_ligand_coord).any(axis=-1)
    ref_ligand_coord  =  ref_ligand_coord[not_nan_mask]
    test_ligand_coord = test_ligand_coord[not_nan_mask]
    # Check if it least one atom is preserved
    assert test_ligand_coord.shape[1] > 0
    rmsd = struc.rmsd(ref_ligand_coord, test_ligand_coord)
    # The deviation of the best pose from the real conformation
    # should be less than 1 Å
    assert rmsd < 1.0

    if flexible:
        # Select best binding pose
        test_receptor_coord = test_receptor_coord[0]
        not_nan_mask = ~np.isnan(test_receptor_coord).any(axis=-1)
        ref_receptor_coord  =  receptor[not_nan_mask]
        test_receptor_coord = test_receptor_coord[not_nan_mask]
        # Check if it least one atom is preserved
        assert test_receptor_coord.shape[1] > 0
        # The flexible residues should have a maximum deviation of 1 Å
        # from the original conformation
        assert np.max(
            struc.distance(test_receptor_coord, ref_receptor_coord)
        ) < 1.0
    else:
        ref_receptor_coord = receptor.coord
        for model_coord in test_receptor_coord:
            assert np.array_equal(model_coord, ref_receptor_coord)
            


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import itertools
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.io as strucio
from ..util import data_dir


# Result should be independent of cell size
@pytest.mark.parametrize("cell_size", [0.5, 1, 2, 5, 10])
def test_get_atoms(cell_size):
    """
    Test the correct functionality of a cell list on a simple test case
    with known solutions.
    """
    array = struc.AtomArray(length=5)
    array.coord = np.array([[0,0,i] for i in range(5)])
    cell_list = struc.CellList(array, cell_size=cell_size)
    assert cell_list.get_atoms(np.array([0,0,0.1]), 1).tolist() == [0,1]
    assert cell_list.get_atoms(np.array([0,0,1.1]), 1).tolist() == [1,2]
    assert cell_list.get_atoms(np.array([0,0,1.1]), 2).tolist() == [0,1,2,3]
    # Multiple positions
    pos = np.array([[0,0,0.1],
                    [0,0,1.1],
                    [0,0,4.1]])
    expected_indices = [0, 1, 2,
                        0, 1, 2, 3,
                        3, 4]
    indices = cell_list.get_atoms(pos, 2)
    assert indices[indices != -1].tolist() == expected_indices
    # Multiple positions and multiple radii
    pos = np.array([[0,0,0.1],
                    [0,0,1.1],
                    [0,0,4.1]])
    rad = np.array([1.0, 2.0, 3.0])
    expected_indices = [0, 1,
                        0, 1, 2, 3,
                        2, 3, 4]
    indices = cell_list.get_atoms(pos, rad)
    assert indices[indices != -1].tolist() == expected_indices


@pytest.mark.parametrize(
    "cell_size, threshold, periodic, use_selection",
    itertools.product(
        [0.5, 1, 2, 5, 10],
        [2, 5, 10],
        [False, True],
        [False, True],
    )
)
def test_adjacency_matrix(cell_size, threshold, periodic, use_selection):
    """
    Compare the construction of an adjacency matrix using a cell list
    and using a computationally expensive but simpler distance matrix.
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    
    if periodic:
        # Create an orthorhombic box
        # with the outer coordinates as bounds
        array.box = np.diag(
            np.max(array.coord, axis=-2) - np.min(array.coord, axis=-2)
        )

    if use_selection:
        np.random.seed(0)
        selection = np.random.choice((False, True), array.array_length())
    else:
        selection = None

    cell_list = struc.CellList(
        array, cell_size=cell_size, periodic=periodic, selection=selection
    )
    test_matrix = cell_list.create_adjacency_matrix(threshold)
    
    length = array.array_length()
    distance = struc.index_distance(
        array,
        np.stack(
            [
                np.repeat(np.arange(length), length),
                  np.tile(np.arange(length), length)
            ],
            axis=-1
        ),
        periodic
    )
    distance = np.reshape(distance, (length, length))
    # Create adjacency matrix from distance matrix
    exp_matrix = (distance <= threshold)
    if use_selection:
        # Set rows and columns to False for filtered out atoms
        exp_matrix[~selection, :] = False
        exp_matrix[:, ~selection] = False
    
    # Both ways to create an adjacency matrix
    # should give the same result
    assert np.array_equal(test_matrix, exp_matrix)


def test_outside_location():
    """
    Test result for location outside any cell.
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    array = array[struc.filter_amino_acids(array)]
    cell_list = struc.CellList(array, cell_size=5)
    outside_coord = np.min(array.coord, axis=0) - 100
    # Expect empty array
    assert len(cell_list.get_atoms(outside_coord, 5)) == 0


def test_selection():
    """
    Test whether the `selection` parameter in the constructor works.
    This is tested by comparing the selection done prior to cell list
    creation with the selection done in the cell list construction.
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    selection = np.array([False, True] * (array.array_length() // 2))
    
    # Selection prior to cell list creation
    selected = array[selection]
    cell_list = struc.CellList(selected, cell_size=10)
    ref_near_atoms = selected[cell_list.get_atoms(array.coord[0], 20.0)]

    # Selection in cell list creation
    cell_list = struc.CellList(array, cell_size=10, selection=selection)
    test_near_atoms = array[cell_list.get_atoms(array.coord[0], 20.0)]

    assert test_near_atoms == ref_near_atoms